/************************************************************

  MD_FCT.c
  
  Multi-dimensional FCT algorithm using Zalesak's generalization.
  Geometry is orthogonal, but not necessarily uniform or rectangular,
  and is defined by struct "grid".

  All fluid quantities are defined on a grid as follows:

 (0,0,...,0,0) (0,0,...,0,1) ..... (0,0,...,0,N[1]+1)
 (0,0,...,1,0) (0,0,...,1,1) ..... (0,0,...,1,N[1]+1)
       .....
 (0,0,...,N[2]+1,0) (0,0,...,N[2]+1,1) ..... (0,0,...,N[2]+1,N[1]+1) 	  

          ....

 (N[M]+1,N[M-1]+1,...,N[2]+1,0)  (N[M]+1,N[M-1]+1,...,N[2]+1,1) ..... 
                                    (N[M]+1,N[M-1]+1,...,N[2]+1,N[1]+1) 


 where N[M] is the size of axis M.

 Externally accessible functions are:

 MD_Flow() : Flows a species set for a specified time interval
 MD_FCT() : Fundamental FCT step as described by Zalesak
 
 External variables are:

 MD_Source_dt : User-modifiable constraint on time step. Useful if 
             source terms are known to require shorter time step than 
	     Courant condition.
 MD_Total_time : Total accumulated time since first call to MD_Flow.

 MD_Courant: Courant condition. Default is .1 Can be changed. 

 MD_FCT_flux_limiter:  weak, strong, medium 
                       corresponding to rhoS, rhoTD, rhoS_init 
		       in Zalesak's eqn (17) and (18)

Revised: 1/16/2018 EAR Fixed accumulation of eps for Courant condition

        Created by Edward A. Richley (richley@mailaps.org)

	This file is subject to the terms and conditions of the 
	GNU General Public License. 

************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifdef ccc
#include <cpml.h>
#endif
#include "basics.h"
#undef pi
#include "MD_FCT_defs.h"
#include "MD_FCT_utils.h"

#define NOT_DONE_WITH_j (!(lim_ctr == N_DIM))

FILE *MD_LOGFILE;
bool MD_trace = false;
MD_STEP MD_current_step = NONE;

static unsigned int *MD_index_steps = NULL;
static unsigned int *MD_index_lim = NULL, N_DIM;
static unsigned int j, lim_ctr, alloc_size, MD_alloc_size, last_version = 0;
static unsigned_int_ptr *MD_index = NULL;
/*static void_fn_ptr *MD_axis_adjust = NULL;*/
static dblptr flux = NULL, rho1 = NULL, nu = NULL, mu = NULL;
static dblptr rhoS_init = NULL, rhoS = NULL, rhoT = NULL,  rhoTD = NULL;
static dblptr  R_plus = NULL, R_minus = NULL; 
static geometry_ptr grid = NULL;
static dblptr  *Volume = NULL, *Area = NULL, *pos = NULL;
static double deltat, Flow_dt, MD_Flow_time;
static string last_name = "";
static MD_FCT_bc_ptr bc;
static MD_FCT_limiter MD_FCT_flux_limiter;

MD_FCT_species_ptr MD_species = NULL;
geometry_ptr MD_grid = NULL;
double MD_Total_time = 0.0, MD_Source_dt, MD_Courant = .1;
double MD_Courant_retry_factor = 0.9999, MD_Courant_update_factor = 1.0;
double MD_FCT_courant_tol = 1.00001;
dblptr MD_FCT_rho_bc;

double MD_FCT();

static void MD_FCT_init_grid_data();
static unsigned int look_at_j_axis();
static unsigned int scan_surface_j();
static void MD_FCT_boundary_conditions();
static double MD_FCT_transport();
static void MD_FCT_get_flux();
static void MD_FCT_diffuse();
static void MD_FCT_flux_corrector();
static void MD_FCT_anti_diffuse();
static double MD_FCT_Lax_Wendroff();
static double MD_FCT_upwind_outflow();
static bool MD_step(MD_FCT_species_set_ptr, bool(), double, MD_FCT_flow_step, bool *, double *);

/* Definition of zero-dimensional grid */

static struct geometry DUMMYGRID = 
{
  "DUMMY",
  1,
  0,
  1,
  0,
  NULL
  };

geometry_ptr DUMMY_GRID = &DUMMYGRID;
/************************************************************

  MD_swap_rho()

  Swaps rho0 and rho1 for all species
  in set. Useful at the end of a full step
  to set up pointers for next step.

************************************************************/
static void MD_swap_rho(set)
MD_FCT_species_set_ptr set;
{
MD_FCT_species_set_rec_ptr  *rec;
MD_FCT_species_ptr sp;

for(rec = &(set->start); (*rec) != NULL; rec = &((*rec)->next)){
  sp = (*rec)->first;
  SWAP_VAL(dblptr, sp->rho0, sp->rho1);
}
return;
}
/************************************************************

  MD_set_ptrs()

  sets pointers for full step.
  set rho to rho1; set rhoS_init to rho0
  for all species in set.

************************************************************/
static void MD_set_ptrs(set)
MD_FCT_species_set_ptr set;
{
MD_FCT_species_set_rec_ptr  *rec;
MD_FCT_species_ptr sp;

for(rec = &(set->start); (*rec) != NULL; rec = &((*rec)->next)){
  sp = (*rec)->first;
  sp->rho = sp->rho1;
  sp->rhoS_init = sp->rho0;
}
return;
}
/************************************************************

  MD_reset_ptrs()

  sets pointers for half step;
  set rho to rho0; set rhoS_init to rho0;

************************************************************/
static void MD_reset_ptrs(set)
MD_FCT_species_set_ptr set;
{
MD_FCT_species_set_rec_ptr  *rec;
MD_FCT_species_ptr sp;

for(rec = &(set->start); (*rec) != NULL; rec = &((*rec)->next)){
  sp = (*rec)->first;
  sp->rho = sp->rho0;
  sp->rhoS_init = sp->rho0;
}
return;
}
/************************************************************

  MD_step()

  Do either a half step or a full step, depending on st. Increment
  MD_Flow_time and MD_Total_time if successful.

  Failure if time step is too big. Global MD_Courant determines the
  maximum allowable eps (v dt/dx). If (max_eps > MD_Courant), calculates
  improved Flow_dt and returns false. Also checks that step size is
  smaller than MD_Source_dt so that any restrictions on step size could be
  computed in user-supplied vel() by having it set MD_Source_dt. 

************************************************************/
static bool MD_step(set, vel, step_size, st, MD_dt_fail_ptr, last_dt)
MD_FCT_species_set_ptr set;
double step_size, *last_dt;
bool (*vel)(), *MD_dt_fail_ptr;
MD_FCT_flow_step st;
{
  MD_FCT_species_set_rec_ptr  *rec;
  MD_FCT_species_ptr sp;
  double src_step_size, flow_step_size, courant, max_eps, h_step, tmp_eps;
  bool succeed;

  h_step = .5*step_size;
  MD_FCT_step = st;
  if (st == half){
    src_step_size = h_step;
    MD_reset_ptrs(set);
    flow_step_size = src_step_size;
    courant = .5*MD_Courant;
  } else {
    src_step_size = step_size;
    MD_set_ptrs(set);
    flow_step_size = step_size;
    courant = MD_Courant;
  }
  *last_dt = src_step_size;
  succeed = vel(set, src_step_size);
  /*
    Test for failure in velocity/source terms.
    Adjust time step limit if fail.
    If succeed, assume MD_Source_dt is valid and
    will take over time step control with Flow_dt
  */
  if (succeed){
    *MD_dt_fail_ptr = false;
  } else {
    *MD_dt_fail_ptr = true;
    return(false);
  }   
/*
  see if source time step violation; may be .5*MD_Source_dt
*/
  if (src_step_size > MD_Source_dt){ 
    return(false); 
  }
  max_eps = 0.0;
  if (MD_trace) { 
    fprintf(MD_LOGFILE,"MD_Total_time: %8.5e flow step: %8.5e %s\n",
	    MD_Total_time, flow_step_size, ((st == half)?"half":"full"));
  }
  for(rec = &(set->start), set->focus = 0; (*rec) != NULL; 
      rec = &((*rec)->next), set->focus++){
    sp = (*rec)->first; 
    tmp_eps = MD_FCT(sp, flow_step_size);
    max_eps = max(max_eps, tmp_eps);
    if (MD_trace){
      fprintf(MD_LOGFILE, "        %s: eps =  %8.5e\n",sp->name, tmp_eps);
    }
  }
  if ((max_eps > MD_FCT_courant_tol*courant)) {
    Flow_dt = flow_step_size*MD_Courant_retry_factor*courant/max_eps;
    return(false);
  }
  MD_Flow_time += h_step;
  MD_Total_time += h_step;
  /* if successful and full step, swap arrays 
     if max_eps is close to zero, increase Flow_dt*/
  if (st == full){ 
    Flow_dt = (max_eps > 0.01)?
      flow_step_size*(MD_Courant_update_factor*MD_Courant/max_eps):2.0*flow_step_size;
    MD_swap_rho(set);
  }
  return(true);
}
/************************************************************

  MD_FCT_test_species_set(set)

  Tests "set" for proper formation of things. Non-null
  boundary conditions offsets, etc.

  What else should we be checking for here? Anything that
  will cause MD_Flow to seg fault with no clues.

************************************************************/
static bool MD_FCT_test_species_set(set)
     MD_FCT_species_set_ptr set;
{
  MD_FCT_species_set_rec_ptr  *rec;
  MD_FCT_species_ptr sp;
  bool result = TRUE, tresult;
  unsigned int i, n_dim;
  MD_FCT_vel_list_rec_ptr vel;

  for(rec = &(set->start); (*rec) != NULL; rec = &((*rec)->next)){
    sp = (*rec)->first;
    n_dim = max(1,sp->grid->num_dim);
    for(i=0; i<n_dim; i++){
      tresult = (bool)(sp->bc[i].OFFSETl == NULL);
      if (tresult == true){
	fprintf(stderr,"MD_FCT_test_species_set: Error in OFFSETl for species: %s\n",
		sp->name);
      }
      result &= !tresult;
      tresult = (bool)(sp->bc[i].OFFSETr == NULL);
      if (tresult == true){
	fprintf(stderr,"MD_FCT_test_species_set: Error in OFFSETl for species: %s\n",
		sp->name);
      }
      result &= !tresult;
    }
    for(vel = sp->vel; vel != NULL; vel = vel->next){
      if (vel->convection != none){
	if (vel->velocity == NULL){
	  fprintf(stderr,"MD_FCT_test_species_set: Error in velocity for axis %s for species %s. No data space.\n",vel->axis->name, sp->name);
	  result &= FALSE;
	}
      }   
    }
  }
  return(result);
}
/************************************************************

  MD_Flow()

  Flows species in set, according to velocities and
  sources defined in user-supplied vel() for a time
  interval Total_Time. Time sequence is a Runge-Kutta
  style two-step method which should be second-order
  accurate in time. Results are in rho0. 

  Performs:

  rhoS = rho0 + .5*dt*Source(rho0)
  rho1 = rhoS + .5*dt*(vel(rho0)*rhoS)/dx
  rhoS = rho0 + dt*Source(rho0)            
  rho1 = rhoS + dt*(vel(rho1)*rhoS)/dx
  rho0 = rho1

  for all species in set until Total_Time is elapsed. 

  User-supplied vel(set, src_step_size) must fill in
  all species' velocities (or re-direct pointers to
  common arrays) based on concentrations in sp->rho.
  (sp->rho is automatically changed to point to
  sp->rho0 or sp->rho1 depending on half-step or
  full-step, so it's best to just use it.). It must 
  also compute rhoS from rhoS_init for all species,
  for an integration time of src_step_size. Note that
  rhoS_init points to either rho0 or rhoS, depending
  on half-step or full-step, so it's important that
  vel() behave properly when (rhoS_init == rhoS) 
  as rhoS is then being over-written (This shouldn't be
  much of a problem in most cases).

************************************************************/
bool MD_Flow(set, vel, Total_Time)
double Total_Time;
bool (*vel)();
MD_FCT_species_set_ptr set;
{
  double start_time, step_size, max_step, total_start_time, last_step_size;
  static double best_guess;
  bool not_done, MD_dt_fail;
  static bool First_Time = true;

  if (MD_FCT_test_species_set(set) == false) {
    fprintf(stderr,"MD_Flow: Errors detected in species set %s\n",set->name);
    fprintf(stderr,"Can not continue\n");
    exit(1);
  }

 if (First_Time){
   best_guess = Total_Time/100.0;
   First_Time = false;
 }

not_done = true;
MD_Flow_time = 0.0;
last_step_size = Flow_dt = best_guess;
MD_Source_dt = Total_Time;
MD_dt_fail = false;

do {
  start_time = MD_Flow_time;
  total_start_time = MD_Total_time;
  max_step = max(0.0, (Total_Time - start_time));
  do {                        /* Runge-Kutta style two-step method */
    do {
      MD_Flow_time = start_time;
      MD_Total_time = total_start_time;
      step_size = min(Flow_dt, .9*MD_Source_dt); /* guard against slowly dropping MD_Source_dt */
      not_done = (step_size < max_step);
      step_size = min(step_size, max_step);
      step_size = (MD_dt_fail)?.5*last_step_size:step_size;
      fprintf(stderr,"step: %lf\n",step_size);
    } while(!MD_step(set, vel, step_size, half, &MD_dt_fail, &last_step_size));
  } while (!MD_step(set, vel, step_size, full, &MD_dt_fail, &last_step_size));
} while(not_done);
best_guess = Flow_dt;
return(false);
}

/************************************************************

  MD_FCT()

  Actual FCT routine for multidimensional flows. 

  Performs:

  rho1 = rhoS + dt*v*rhoS/dx

 Assumes spe->vel is provided, having velocities in all directions
 specified.  Also uses spe->rhoS to allow source terms to be included.
 Allows rhoS to point to rho0 if appropriate (doesn't modify rho0 but
 does apply boundary conditions to rhoS. rho0 is not used. rhoS
 and rhoTD are used for flux correction purposes.

 If spe has a zero-dimensional grid (DUMMY_GRID, for example),
 spe->rho1 is set to spe->rhoS, and a value of zero is returned.

************************************************************/
double MD_FCT(spe, dt )
MD_FCT_species_ptr spe;
double dt;
{
unsigned int i;
static unsigned int flux_alloc = 0;
double result;

MD_current_step = MD_FCT_STEP;
MD_species = spe;
MD_FCT_flux_limiter = spe->Flux_Limiter;
MD_grid = grid = spe->grid;
if (grid->num_dim == 0){ 
  for(i=0; i<grid->alloc_size; i++){
    spe->rho1[i] = spe->rhoS[i];
  }
  return(0.0);
}
MD_FCT_init_grid_data();
deltat = dt;
if ((flux == NULL)|(flux_alloc < MD_alloc_size)){
  nu = get_memory(nu, MD_alloc_size, double);
  mu = get_memory(mu, MD_alloc_size, double);
  R_plus = get_memory(R_plus, alloc_size, double);
  R_minus = get_memory(R_minus, alloc_size, double);
  if ((nu == NULL)|(mu == NULL)|
      (R_plus == NULL)|(R_minus == NULL)){
    fprintf(stderr,"ERROR: Insufficient memory [MD_FCT]\n");
    exit(1);
  }
  flux_alloc = MD_alloc_size;
}
flux = mu;
rhoS_init = spe->rhoS_init;
rhoS = spe->rhoS;
rho1 = spe->rho1;
rhoTD = rho1;
rhoT = R_plus;
bc = spe->bc;
MD_FCT_boundary_conditions(rhoS);
result = MD_FCT_transport();                     /* rhoS ==> rhoT */ 

MD_FCT_boundary_conditions(rhoT);
MD_FCT_diffuse();                       /* rhoT ==> rhoTD */
MD_FCT_boundary_conditions(rhoTD);
MD_FCT_get_flux();                     /* flux(rhoT) */
MD_FCT_flux_corrector(spe); /* fluxC(rhoS, rhoTD), uses R+, R- for scratch */
MD_FCT_anti_diffuse();                 /* rhoTD ==> rho1 */
MD_FCT_boundary_conditions(rho1);
return(result);
}
/************************************************************

  MD_FCT_boundary_conditions()

  sets boundary conditions on array rho on grid to 
  those specified by pointer bc. Method is slope
  and intercept as described in Boris1993. bc points
  to an array of MD_FCT_bc entries; one for each axis.
  Conditions can be local or periodic. In either case:

  rho[boundary] = S*rho[location] + V +
                  (rhoS + dt*vchar*drhoS/dx)

  where location is either the nearest location (local) or
  the one at the opposite end of the axis (periodic) and
  where the outflow term is evaluated only if Lchar and vchar
  are not zero, and is represented by a Lax-Wendroff step.

  This is all over-ridden by (bc.constant == true), which will
  just leave the boundary points alone and use them. 

************************************************************/
static void MD_FCT_boundary_conditions(rho)
dblptr rho;
{
unsigned int sl, sr, sld, srd, ssteps, tindex;
double dr, t_rho;
axis_ptr axs;

MD_FCT_rho_bc = rho;
for(axs = MD_grid->axes,j=0; j<N_DIM; axs = axs->next, j++){     /* looking at 
the j axis */
  sl = look_at_j_axis();
  ssteps = MD_index_steps[j];
  do{
    sl = scan_surface_j(sl);
    sr = sl + ssteps*MD_index_lim[j];
    sld = sl + (*(bc[j].OFFSETl))*ssteps;
    srd = sr - (*(bc[j].OFFSETr))*ssteps;
/* Set the boundary conditions ... first the left 
  if a function has been defined, set axes index
  to point to just inside the boundary	  */
    tindex = *MD_index[j];
    if (bc[j].Fl != NULL){bc[j].Fl(bc+j);}
    if (bc[j].constant_L) {
      rho[sl] = rhoS[sl];
    } else {
      t_rho = bc[j].Sl*rho[sld]+bc[j].Vl;
      /* only do Lax-Wendroff step if Lchar!= 0.0, AND it's 
	 the transport step */
      if (bc[j].Lchar_l != 0.0){
	if (rho == rhoS) {
	  dr = (pos[j][sl+ssteps] - pos[j][sl]);
	  t_rho += MD_FCT_Lax_Wendroff(rhoS[sl], rhoS[sld], bc[j].rho_amb_l, 
				       dr, bc[j].Lchar_l, 
				       bc[j].vchar_l, deltat);
	} else {
	  t_rho += rhoS[sl];
	}
      }
      rho[sl] = t_rho;
    }
/* then the right */
    *MD_index[j] = MD_index_lim[j];
    if (bc[j].Fr != NULL){bc[j].Fr(bc+j);}
    if (bc[j].constant_R) {
      rho[sr] = rhoS[sr];
    } else {
      t_rho = bc[j].Sr*rho[srd]+bc[j].Vr;
      if (bc[j].Lchar_r != 0.0){
	if (rho == rhoS){
	  dr = (pos[j][sr] - pos[j][sr-ssteps]);
	  t_rho += MD_FCT_Lax_Wendroff(rhoS[sr], rhoS[srd], bc[j].rho_amb_r, 
				       dr, bc[j].Lchar_r, bc[j].vchar_r, deltat);
	} else {
	  t_rho += rhoS[sr];
	}
      }
      rho[sr] = t_rho;
    }
    *MD_index[j] = tindex;
  } while (NOT_DONE_WITH_j);
}
return;
}
/************************************************************

  MD_FCT_transport()

  transport step depends on conv as follows:
  
  conv:         performs:
  conservation  rhoT = rhoS + deltat*deldot(rhoS*vel)
                       with rhoS and vel interpolated
		       by arithmetic mean
  geometric_conservation  
                rhoT = rhoS + deltat*deldot(rhoS*vel)
		       with rhoS interpolated by 
		       geometric mean for anti-terracing
  flux_conservation
                rhoT = rhoS + deltat*deldot(rhoS*vel)
		       with (rhoS*vel) interpolated 
		       by arithmetic mean
  geometric_flux_conservation
                rhoT = rhoS + deltat*deldot(rhoS*vel)
		       with (rhoS*vel) interpolated 
		       by geometric mean
  advection     rhoT = rhoS + deltat*vel*grad(rhoS)
  zip           rhoT = rhoS + deltat*deldot(rhoS*vel) 
                       with Hirt's "zip" fluxes
  compression   rhoT = rhoS + deltat*rhoS*deldot(vel)		       
  none          rhoT = rhoS

************************************************************/
static double MD_FCT_transport()
{
  unsigned int i, sl, vl, s, v, s_next, v_next, ssteps, n_vel, p, cnt;
  double Vol_i, Vol_next, Vol_mid, x1, x2, x;
  double vel_i, one_Vol_i, vel_next, vel_mid, one_Vol_next, one_Vol_mid;
  double ABS_eps_over_2, eps_over_2, eps2;
  double a, a_last, rhoS_i, rhoS_next, *rhox, result, rhoS_last;
  double *Vj, *Aj, *vel; 
  MD_FCT_convection cj;
  bool skip, First, execute, accumulate;
  MD_FCT_vel_list_rec_ptr ivel, tvel;

  First = true;
/***** Scan axes, vl is base pointer for nu and mu quantities *****/
  result = 0.0; a_last = 0.0;
  /*
    for each axis, find out how many velocities there are, so that
    the last one can be used to obtain nu and mu. n_vel is the number
    of velocities on this axis. If MD_species->vel == NULL, then n_vel == 0.
  */
  for(j = vl = 0; j < N_DIM; j++, vl += alloc_size){
    for(p = 0, ivel = MD_species->vel; ivel != NULL; ivel = ivel->next) {
      if (ivel->axis->sequence == j) {p++;}
    }; 
    n_vel = p;

    if (First){
      rhox = rhoS;
      First = false;
    } else {
      rhox = rhoT;
    }
    sl = look_at_j_axis();
    /* 
       make sure you do it at least once per axis; 
       then as many times as there are velocities on that axis 
    */
    for(cnt = 0, p = 0, ivel = MD_species->vel; (ivel != NULL)||(cnt == 0);  cnt++) {
      execute = false;
      if (MD_species->vel == NULL) {
	cj = none;
	execute = true;
	tvel = ivel;
      } else {
	if (ivel->axis->sequence == j) {
	  p++; 
	  execute = true;
	  cj = ivel->convection;
	  tvel = ivel;
	}
	vel = ivel->velocity;
	ivel = ivel->next;
      }
      if (execute){
	accumulate =  (p != n_vel);
	ssteps = MD_index_steps[j];
	skip = (cj == none);
	Vj = Volume[j];
	Aj = Area[j];
	//	result = 0.0; // Changed 1/16/2018 EAR. Not sure why this is here when it's above. 
	do{ 
	  sl = scan_surface_j(sl);
	  v = (vl+sl);
	  vel_i = skip?0.0:vel[sl];
	  Vol_i = Vj[sl];
	  one_Vol_i = 1.0/Vol_i;
	  rhoS_last = (rhoS_i = rhoS[sl]);
	  for(s = sl, i = 0;
	      i<MD_index_lim[j]; s = s_next, i++){ /* go down each axis */
	    s_next = s + ssteps;
	    v_next = vl + s_next;
	    rhoS_next = rhoS[s_next];
	    Vol_next = Vj[s_next];
	    one_Vol_next =  1.0/Vol_next;
	    Vol_mid = .5*(Vol_i + Vol_next);
	    one_Vol_mid = .5*(one_Vol_i + one_Vol_next); 
	    if (skip) {
	      eps_over_2 = vel_next = 0.0;
	    } else {
	      vel_next = vel[s_next];
	      vel_mid = .25*(vel_next + vel_i)*Aj[s]*deltat;
	      eps_over_2 = vel_mid*one_Vol_mid;
	    }
	    /* Note: mu[v] accumulates eps_over_2 for all velocities on a given axis */
	    if (accumulate){
	      if (p == 1) {
		mu[v] = eps_over_2;
	      } else {
		mu[v] += eps_over_2;
	      }
	    } else {
	      if (n_vel > 1){
		eps_over_2 += mu[v];
	      }
	      ABS_eps_over_2 = ABS(eps_over_2);
	      result = max(result, ABS_eps_over_2);
	      eps2 = TWO_THIRDS*eps_over_2*eps_over_2;
	      mu[v] = Vol_mid*(ONE_SIXTH - eps2);
	      nu[v] = Vol_mid*(ONE_SIXTH + eps2 + eps2)*
		(rhoS_next - rhoS_i);
	    }
	    switch (cj) {
	    case conservation: 
	      a = (rhoS_next+rhoS_i)*vel_mid;
	      if (i) {
		rhoT[s] = rhox[s] + (a_last - a)*one_Vol_i;
	      }
	      a_last = a;
	      break;
	    case geometric_conservation:
	      x = rhoS_next*rhoS_i;
	      x = (x >= 0.0)?(2.0*sqrt(x)):(rhoS_next+rhoS_i);
	      a = x*vel_mid;
	      if (i) {
		rhoT[s] = rhox[s] + (a_last - a)*one_Vol_i;
	      }
	      a_last = a;
	      break;
	    case flux_conservation:
	      a = .5*deltat*(rhoS_next*vel_next + rhoS_i*vel_i)*Aj[s];
	      if (i) {
		rhoT[s] = rhox[s] + (a_last - a)*one_Vol_i;
	      }
	      a_last = a;
	      break;
	    case geometric_flux_conservation:
	      x1 = rhoS_next*vel_next;	    x2 = rhoS_i*vel_i;
	      x = x1*x2;
	      a = deltat*Aj[s]*((x>=0.0)?sqrt(x):.5*(x1+x2));
	      if (i) {
		rhoT[s] = rhox[s] + (a_last - a)*one_Vol_i;
	      }
	      a_last = a;
	      break;
	    case advection:
	      if (i) {
		rhoT[s] = rhox[s] - vel_i*deltat*
		  (rhoS_next - rhoS_last)/(pos[j][s_next] - pos[j][s-ssteps]);
		rhoS_last = rhoS_i;
	      }
	      break;
	    case zip:
	      a = .5*deltat*
		(rhoS_next*vel_i + rhoS_i*vel_next)*Aj[s];
	      if (i) {
		rhoT[s] = rhox[s] + (a_last - a)*one_Vol_i;
	      }
	      a_last = a;
	      break;
	    case compression:
	      a = vel_mid;
	      if (i) {
		rhoT[s] = rhox[s] + rhoS_i*(a_last -a)*one_Vol_i;
	      }
	      a_last = a;
	      break;
	    case none:
	      if (i){
		rhoT[s] = rhox[s];
	      }
	      break;	  
	    } 
	    vel_i = vel_next;
	    rhoS_i = rhoS_next;
	    Vol_i = Vol_next;
	    one_Vol_i = one_Vol_next;
	    v = v_next;
	  }
	} while(NOT_DONE_WITH_j);
      }
    }
  }
  return(2.0*result);
}
/************************************************************

  MD_FCT_diffuse()

  diffusion step performs:
  
  rhoTD = rhoT + Sum(j){(nu(i,j) - nu(i-1,j))}

  where nu() has already been determined as the
  diffusion flux.

************************************************************/
static void MD_FCT_diffuse()
{
  unsigned int i, sl, vl, s, v, limit, ssteps;
  double *rhox, nu_i, nu_last, *Vj;
/*****  Loop over axes. vl is base pointer for vector quantities *******/
  nu_last = 0.0;
  for(j = vl = 0 ; j < N_DIM; j++, vl += alloc_size){
    limit = MD_index_lim[j];
    ssteps = MD_index_steps[j];
    rhox = j?rhoTD:rhoT;
    sl = look_at_j_axis();
    Vj = Volume[j];
    do{ 
      sl = scan_surface_j(sl);
      for(s = sl, i = 0; i<limit; s += ssteps, i++){ /* go down each axis */
	v = vl + s;
	nu_i = nu[v];
	if (i) {
	  rhoTD[s] = rhox[s] + (nu_i - nu_last)/Vj[s];
	}
	nu_last = nu_i;
      } 
    } while(NOT_DONE_WITH_j);
  } 
  return;
}
/************************************************************

  MD_FCT_get_flux()

  Gets uncorrected anti-diffusion fluxes:
  
  flux(i,j) = mu(i,j)*(rhoT(i+1) - rhoT(i))

  and do preliminary correction by comparing to slope of rhoTD

************************************************************/
static void MD_FCT_get_flux()
{
  unsigned int i, sl, sr, vl, s, v, s_next, limit, v_last, ssteps, O_L, O_R;
  double rhoT_i, rhoT_next, fl, rhobl, rhobr;
  bool b1, b2, b3, b5, b6, b7, b8, Per_l, Per_r;
/*****  Loop over axes. vl is base pointer for vector quantities *******/
  b1 = b5 = b7 = false;
  v_last = 0; fl = 0.0;
  for(j = vl = 0;  j < N_DIM; j++, vl += alloc_size){
    sl = look_at_j_axis();
    limit = MD_index_lim[j];
    ssteps = MD_index_steps[j];
    do{ 
      sl = scan_surface_j(sl);
      sr = sl + limit*ssteps;
      rhoT_i = rhoT[sl];
      v = vl + sl;
      b2 = false; rhobl = rhobr = 0.0;
/* drop the conditions which are two cells off the grid 
   unless it's periodic boundary conditions */
      Per_l = ((O_L = *(bc[j].OFFSETl)) != 1);
      if (Per_l) {
	rhobl = rhoTD[sl+ssteps*O_L];
	b2 = (rhoTD[sl] > rhobl);
      }
      Per_r = ((O_R = *(bc[j].OFFSETr)) != 1);
      if (Per_r) {
	rhobr = rhoTD[sr-ssteps*O_R];
      }
      for(s = sl, i = 0;
	  (i<limit); s = s_next, i++){ /* go down each axis */
	s_next = s + ssteps;
	rhoT_next = rhoT[s_next];
	v = vl + s;
	b3 = (rhoTD[s_next] > rhoTD[s]);
	b6 = (b1 != b3);
	if (i) {	
	  b8 = b5&&(b6||b7);
	  flux[v_last] = b8?0.0:fl; 
	} 
	fl = mu[v]*(rhoT_next-rhoT_i);
	b1 = (fl > 0.0);
	b7 = (b1 != b2)&&(i||Per_l);
	b2 = b3;
	b5 = (b1 != b2);
	v_last = v;
	rhoT_i = rhoT_next;
      } 
      b6 = Per_r?(b1 != (b3 = (rhobr > rhoTD[sr]))):false;
      b8 = b5&&(b7||b6);
      flux[v_last] = b8?0.0:fl;
    } while(NOT_DONE_WITH_j);
  }
  return;
}
/************************************************************

  MD_FCT_flux_corrector()

  As in Zalesak1979, Oran1987 sec. 8-4. RhoS and RhoTD are 
  used to set rho_max and rho_min.

************************************************************/
static void MD_FCT_flux_corrector(sp)
MD_FCT_species_ptr sp;
{
  unsigned int i, sl, sr, v_lo, v_hi, s, v, s_lo, s_hi, limit, ssteps;
  unsigned int p, psteps, vl, s_next, Offsetl, Offsetr;
  double P_plus, P_minus, rho_max, rho_min, f_lo, f_hi;
  double one_Vol_i, C, m1, m2, sp_max, sp_min; 
  bool enable_max, enable_min;
  dblptr rhox;

  enable_max = sp->enable_maximum;
  enable_min = sp->enable_minimum;
  sp_max = sp->maximum;
  sp_min = sp->minimum;
/*
  Get all the R_plus, R_minus entries
*/
  switch(MD_FCT_flux_limiter)
    {
    case weak:
      rhox = rhoS;
      break;
    case strong:
      rhox = rhoTD;
      break;
    case medium:
      rhox = rhoS_init;
      break;
    }
  j = 0;
  limit = MD_index_lim[0];
  ssteps = MD_index_steps[0];
  sl = look_at_j_axis();
  do{ 
    sl = scan_surface_j(sl);
    for(s = sl+ssteps, i = 1; i<limit; s += ssteps, i++){ /* scan each entry */
      P_plus = (P_minus = 0.0);
      rho_max = max(rhox[s], rhoTD[s]);
      rho_min = min(rhox[s], rhoTD[s]);
      for(p=0; p< N_DIM; p++){          /* Look along each axis */
	one_Vol_i = 1.0/Volume[p][s];
	psteps = MD_index_steps[p];
	s_hi = s + psteps;
	s_lo = s - psteps;
	v_hi = p*alloc_size + s;
	f_hi = flux[v_hi];
	v_lo = v_hi - psteps;
	f_lo = flux[v_lo];
	P_plus += (max(0.0, f_lo) - min(0.0, f_hi))*one_Vol_i;
	P_minus += (max(0.0, f_hi) - min(0.0, f_lo))*one_Vol_i;
	m1 = max(rhox[s_hi],rhoTD[s_hi]);
	m2 = max(rhox[s_lo],rhoTD[s_lo]);
	m2 = max(m1,m2);
	rho_max = max(rho_max, m2);
	m1 = min(rhox[s_lo], rhoTD[s_lo]);
	m2 = min(rhox[s_hi], rhoTD[s_hi]);
	m2 = min(m1, m2);
	rho_min = min(rho_min, m2);
      }
      rho_max = enable_max?min(rho_max, sp_max):rho_max;
      rho_min = enable_min?max(sp_min,rho_min):rho_min;

      m1 = (P_plus == 0.0)?0.0:((rho_max - rhoTD[s]))/P_plus;
      m2 = (P_minus == 0.0)?0.0:((rhoTD[s] - rho_min))/P_minus;
      R_plus[s] = min(1.0,m1);
      R_minus[s] = min(1.0,m2);
    }
  } while(NOT_DONE_WITH_j);
/*
  Scan all the flux entries, correct each one
*/
  for(j = vl = 0; j < N_DIM; j++, vl += alloc_size){
    sl = look_at_j_axis();
    limit = MD_index_lim[j];
    ssteps = MD_index_steps[j];
    do{ 
      sl = scan_surface_j(sl);
      sr = sl + limit*ssteps;
/*
  The boundary values of R_plus and R_minus have to be fixed up.
  Equal to local values by default, made equal to opposite
  values if periodic.
*/
      Offsetl = *(bc[j].OFFSETl);
      if (Offsetl == 1) {
	R_plus[sl] = R_minus[sl+ssteps];
	R_minus[sl] = R_plus[sl+ssteps];
      } else {
	R_plus[sl] = R_plus[sl+ssteps*(Offsetl)];
	R_minus[sl] = R_minus[sl+ssteps*(Offsetl)];
      }
      Offsetr = *(bc[j].OFFSETr);
      if (Offsetr == 1) {
	R_plus[sr] = R_minus[sr-ssteps];
	R_minus[sr] = R_plus[sr-ssteps];
      } else {
	R_plus[sr] = R_plus[sr - ssteps*(Offsetr)];
	R_minus[sr] = R_minus[sr - ssteps*(Offsetr)];
      }
      for(s = sl, i = 0; (i<limit); s = s_next, i++){ /* go down each axis */
	s_next = s + ssteps;
	v = vl + s;
	C = (flux[v] > 0.0)?
	  min(R_plus[s_next],R_minus[s]):
	    min(R_plus[s],R_minus[s_next]);
	flux[v] *= C;	
      }
    } while(NOT_DONE_WITH_j);
  } 
}

/************************************************************

  MD_FCT_anti_diffuse()

  Performs anti-diffusion with corrected fluxes in flux[].

  Performs:

  rho1 = rhoTD + Sum(j){(flux[j][i-1] - flux[j][i])/Volume[j]/[i]}

************************************************************/
static void MD_FCT_anti_diffuse()
{
  unsigned int i, sl, vl, s, v, limit, ssteps;
  double *rhox, fl_i, fl_last, *Vj;
/*****  Loop over axes. vl is base pointer for vector quantities *******/
  fl_last = 0.0;
  for(j = vl = 0; j < N_DIM; j++, vl += alloc_size){
    rhox = j?rho1:rhoTD;
    limit = MD_index_lim[j];
    ssteps = MD_index_steps[j];
    sl = look_at_j_axis();
    Vj = Volume[j];
    do{ 
      sl = scan_surface_j(sl);
      for(s = sl, i = 0; i<limit; s += ssteps, i++){ /* go down each axis */
	v = vl + s;
	fl_i = flux[v];
	if (i) {
	  rho1[s] = rhox[s] + (fl_last - fl_i)/Vj[s];
	}
	fl_last = fl_i;
      } 
    } while(NOT_DONE_WITH_j);
  } 
return;
}

/************************************************************

  MD_FCT_init_grid_data()

  sets up arrays for indexing through grid.
  Used by MD_FCT and by MD_FCT_print_species

************************************************************/
static void MD_FCT_init_grid_data()
{
unsigned int i;
axis_ptr axs;

N_DIM = grid->num_dim;
alloc_size = grid->alloc_size;
MD_alloc_size = N_DIM*alloc_size;
/*
  see if grid has changed, or if more space is needed
*/
if (strcmp(last_name, grid->name)||(grid->version != last_version)){
  MD_index = get_memory(MD_index, N_DIM, unsigned_int_ptr);
  MD_index_lim = get_memory(MD_index_lim, N_DIM, unsigned int);
  MD_index_steps = get_memory(MD_index_steps, N_DIM, unsigned int);
  Volume = get_memory(Volume, N_DIM, dblptr);
  Area = get_memory(Area, N_DIM, dblptr);
  pos = get_memory(pos, N_DIM, dblptr);
  for(axs = (grid->axes), i=0; axs != NULL; axs = (axs->next),i++){
    MD_index_lim[i] = (axs->num_pts+1);
    MD_index_steps[i] = i?(MD_index_steps[i-1]*(MD_index_lim[i-1]+1)):1;
    Volume[i] = axs->Volume;
    Area[i] = axs->Area;
    pos[i] = axs->r;
    MD_index[i] = &(axs->index);
  }
  last_version = grid->version;
  last_name = grid->name;
}
return;
}

/************************************************************

  look_at_j_axis()

  find the array index for beginning of axis pointed to by global j,
  and set up MD_index entries to point there (actually one
  step before the beginning)

************************************************************/
static unsigned int look_at_j_axis(){
  unsigned int result, i;
  result = 0;
  lim_ctr = N_DIM;                  /* looking at the j axis */
  for(i=0; i<N_DIM; i++){  /* set index counter to (num_pts, num_pts, ..., 
num_pts) */
    *(MD_index[i]) = MD_index_lim[i]-1;
    result += (i==j)?0:(*(MD_index[i]))*MD_index_steps[i];
  }
  return(result);
}
/************************************************************

  scan_surface_j()

  finds increment to point to next vector along axis j
   determine how many axes are at their limits, and 
   get increment in array index for left side of next vector along 
   axis j 


************************************************************/
static unsigned int scan_surface_j(s)
unsigned int s;
{
  unsigned int i,p;
  static unsigned_int_ptr ind;
  for(i= !j; (i<N_DIM); i++, i+=(j==i)){ 
    ind = MD_index[i];
    *ind = ((p = (*ind + 1)) ==  MD_index_lim[i])?1:p; 
    lim_ctr += (p == MD_index_lim[i]-1); 
    if (*ind == 1) {  
      s -= (MD_index_lim[i]-2)*MD_index_steps[i]; 
      lim_ctr--; 
    } else { 
      s += MD_index_steps[i]; 
      break; 
    } 
  } 
  /*  if (*(MD_axis_adjust[j]) != NULL){
    (*MD_axis_adjust[j])();
  }*/
  return(s);
}
/************************************************************

  MD_Diffusion

  Implicit, one-dimensional diffusion step for implementing
  parabolic terms of the form:

  d (rho)
  ______  = div( D * grad(kappa*rho))
   dt

  in the MD_FCT context. Intended for use with fractional step
  coupling. Based on a tri-diagonal implicit step. Does not work
  with periodic bc because that system it not tri-diagonal.

  spe:      species to diffuse

  dt:       time step over which to diffuse

  ax:       axis pointer on which to apply diffusion

  Dfn:      function to return D and kappa

  switches: byte used for performance optimization and control.
            made up of flags:
	    MD_Diffusion_re_use_Dkappa  re-use diffusion matrix from
	                           previous call, as if D and kappa
				   are time-invariant. saves
				   computation time

	    MD_Diffusion_first_fractional_step take spe->rho_n
	                           to spe->rho_np1. Otherwise,
				   take spe->rho_np1 to
				   spe->rho_np1.

	    note: Also, (A = I - dt*D), so that A is re-used
	          (in a multi-dimensional problem)
		  if D and dt haven't changed (if D is
		  constant and it's not the first sweep
		  through the grid). A is always generated
		  anew for the first sweep (assumes dt
		  is different from last call).

		  If boundary conditions are simple (L_char == 0.0), 
		  then S*rho+V methods are used. If not simple, 
		  treated as if ambient conditions exist at a distance
		  of L_char. Any existing bc functions are called. 

  returns maximum (D*kappa*Area/(dr*Volume)) for time step control.

************************************************************/
double MD_Diffusion(spe, dt, ax, Dfn, switches)
     MD_FCT_species_ptr spe;
     double dt;
     axis_ptr ax;
     MD_diffusivity (*Dfn)();
     char switches;
{
  bool try_to_re_use_D, re_use_D, re_use_A, first_step;
  bool bc_l_simple, bc_r_simple, First_Sweep;
  unsigned int sl, sr, N, i, tmp_ind, ssteps, p, Npts, tindex, end_step, s;
  dblptr rho_in, rho_out, Aj, Vj;
  double one_Vol, D_mid_l, r, dr, r_last, db;
  double rho_in_l, rho_in_r;
  double L_char_l, L_char_r, kappa_l, kappa_r, D_l, D_r;
  MD_diffusivity Dk;
  static dblptr A = NULL, D = NULL, Da, Db, Dc, Aa, Ab, Ac;
  static dblptr Di, kappai, D_mid, Daddr, kaddr;
  static unsigned int N_last = 0;
  static double max_eps = 0.0;

  MD_current_step = MD_DIFFUSION_STEP;
  MD_species = spe;
  bc = spe->bc;
  grid = spe->grid;
  MD_FCT_init_grid_data();
  j = ax->sequence;

  Npts = (N = MD_index_lim[j]-1)+2;

  try_to_re_use_D = (switches&MD_Diffusion_re_use_Dkappa);
  re_use_D = try_to_re_use_D;

  if (Npts > N_last){
    A = get_memory(A, (3*Npts), double);
    Aa = A; Ab = A+Npts; Ac = Ab+Npts;
    Di = Aa; kappai = Ab; D_mid = Ac;
    N_last = Npts;
  }
  D = spe->D[j];
  if (D == NULL){
    D = get_memory(D, (3*Npts), double);
    re_use_D = false;
    spe->D[j] = D;
  }
  Da = D; Db = Da+Npts; Dc = Db+Npts;
  first_step = switches&MD_Diffusion_first_fractional_step;
  rho_in = first_step?(spe->rho_n):(spe->rho_np1);
  rho_out = spe->rho_np1;
  MD_FCT_rho_bc = rho_in;
  ssteps = MD_index_steps[j];
  Aj = Area[j]; Vj = Volume[j];
  ssteps = MD_index_steps[j];
  end_step = ssteps*MD_index_lim[j];
  sl = look_at_j_axis();
  max_eps = re_use_D?max_eps:0.0;
  First_Sweep = true;
  do{
    sl = scan_surface_j(sl);
    sr = sl + end_step;
    rho_in_l = rho_in[sl]; rho_in_r = rho_in[sr];
    if (bc[j].Fl != NULL) {bc[j].Fl(bc+j);}
    bc_l_simple = ((L_char_l = (bc[j].Lchar_l)) == 0.0);
    tindex = *(MD_index[j]);
    *(MD_index[j]) = MD_index_lim[j];
    if (bc[j].Fr != NULL) {bc[j].Fr(bc+j);}
    *(MD_index[j]) = tindex;
    bc_r_simple = ((L_char_r = (bc[j].Lchar_r)) == 0.0);
    /**************************************************
      
      get new D matrix if necessary. OK to use A as storage   
      space (with Di, kappai, D_mid) since new D implies new A 
      
      ***************************************************/
    if (!re_use_D) { 
      tmp_ind = *(MD_index[j]);
      Daddr = &D_l; kaddr = &kappa_l;
      r = ax->r[sl] - L_char_l; dr = 0.0;
      for(i=0, p = 0, s = sl; ; p = i++, s+=ssteps){
	/*	Dk = Dfn(grid, r);*/
	Dk = Dfn(spe, s, r);
	*Daddr = Dk.D; *kaddr = Dk.kappa;
	if(p){
	  D_mid[p-1] = ((!bc_r_simple)||(p != Npts))?
	    (.5*(*Daddr + Di[p-1])*Aj[(p == Npts)?(s-ssteps-ssteps):(s-ssteps)]/dr):0.0;
	} else if (i){
	  D_mid_l = bc_l_simple?0.0:(.5*(Di[0] + D_l)*Aj[sl]/L_char_l);
	}
	if (i > Npts) {break;}
	if (i == Npts){
	  r += L_char_r;
	  Daddr = &D_r;  kaddr = &kappa_r;
	  dr = L_char_r;
	} else {
	  r_last = r;
	  dr = (r = ax->r[s]) - r_last;
	  Daddr = (Di+i);  kaddr = (kappai+i);
	  *(MD_index[j]) = min(i,(Npts-1));
	}
      }
      for(i=1, p=0, s=sl+ssteps; i <= N; p = i++, s+=ssteps){
	one_Vol = 1.0/Vj[s];
	Da[i] = D_mid[p]*kappai[p]*one_Vol;
	Db[i] = -(db = (D_mid[i]+D_mid[p])*kappai[i]*one_Vol);
	Dc[i] = D_mid[i]*kappai[i+1]*one_Vol;
	max_eps = max(max_eps, db);
      }
      if (bc_l_simple){
	Da[0] = Db[0] = Dc[0] = 0.0;
      } else {
	Da[0] = D_mid_l*kappa_l/Vj[sl];
	Db[0] = -(D_mid[0]+D_mid_l)*kappai[0]/Vj[sl];
	Dc[0] = D_mid[0]*kappai[1]/Vj[sl];
      }
      if(bc_r_simple){
	Da[N+1] = Db[N+1] = Dc[N+1] = 0.0;
      } else {
	Da[N+1] = D_mid[N]*kappai[N]/Vj[sr];
	Db[N+1] = -(D_mid[N+1]+D_mid[N])*kappai[N+1]/Vj[sr];
	Dc[N+1] = D_mid[N+1]*kappa_r/Vj[sr];
      }
      *(MD_index[j]) = tmp_ind;
    }
    /**************************************************

      if A can't be re-used, compute new values

      **************************************************/
    re_use_A = re_use_D&&(!First_Sweep);
    if (!re_use_A) {
      for(i=0; i<Npts; i++){
	Aa[i] = -dt*Da[i];
	Ab[i] = 1.0-dt*Db[i];
	Ac[i] = -dt*Dc[i];
      }
    }
    re_use_D = try_to_re_use_D;
    /**************************************************
      
      Fix up boundary conditions

      **************************************************/
    if (bc[j].constant_L) {
      Ac[0] = 0.0;
    } else {
      if (bc_l_simple){
	rho_in[sl] = bc[j].Vl;
	Ac[0] = - bc[j].Sl;
      } else {
	Ac[0] -= bc[j].Sl;
	rho_in[sl] += dt*Da[0]*bc[j].rho_amb_l + bc[j].Vl;
      }
    }
    if (bc[j].constant_R){
      Aa[N+1] = 0.0;
    } else {
      if (bc_r_simple){
	rho_in[sr] = bc[j].Vr;
	Aa[N+1] = - bc[j].Sr;
      } else {
	Aa[N+1] -= bc[j].Sr;
	rho_in[sr] += dt*Dc[N+1]*bc[j].rho_amb_r + bc[j].Vr;
      }
    }
    /**************************************************

      Solve the matrix

      **************************************************/
    if (!MD_tri_diag(A, &(rho_out[sl]), &(rho_in[sl]), N, ssteps, 
(!re_use_A))){
      fprintf(stderr,"[MD_Diffusion]: singular A matrix\n");
      exit(1);
    }
    rho_in[sl] = rho_in_l; rho_in[sr] = rho_in_r;
    First_Sweep = false;
  } while(NOT_DONE_WITH_j);
  return(.5*max_eps);
}
/************************************************************

  MD_FCT_Lax_Wendroff(Rho1, Rho2, Rho3, dx1, dx2, V, dt)

  performs one modified Lax-Wendroff step at constant velocity, V
  for time dt. For outflow boundary conditions. 
  
  Performs:

  d(RhoG)/dt = - V*d(Rho)/dx - (V/L)(RhoG - RhoAmb)

  d(RhoG)/dt = - V*d(Rho)/dx 

  where V is assumed positive (characteristic speed away from the
  grid), L is a characteristic length over which RhoG decays to
  RhoAmb. 

  A Lax-Wendroff term is added to the above, and it is all evaluated
  semi-implicitly. Implicitness is important because V may be much
  larger than flow velocities in subsonic problems, thus causing 
  (vdt/dr >> 1).  Thus...transition to extrapolated values is automatic.

  *****  Something is VERY WRONG with this formula. MUST FIX.*******

  RhoN is the value just inside the computational grid. RhoG is the
  value just before the bc is applied. RhoAmb is the ambient value
  at distance L. dr is the grid spacing between guard cell and last
  computational cell. V is the characteristic speed (assumed positive).
  dt is the time step.

  Used only for outflow boundary conditions.

************************************************************/
static double MD_FCT_Lax_Wendroff(RhoG, RhoN, RhoAmb, dr, L, V, dt)
double RhoG, RhoN, RhoAmb, dr, L, V, dt;
{
double x, y, z, w, v, result;

x = (V*dt); y = x*(dr + x); z = L+dr;
v = dr*L; w = v*z; 

result = (w*RhoG + y*(L*RhoN + dr*RhoAmb))/(w + y*z);
/*result = (w*RhoG + x*(v*(RhoN-RhoAmb) + x*(L*RhoN + dr*RhoAmb)))/(w + 
x*x*z);*/
return(result);


}
/************************************************************

  MD_FCT_upwind_outflow(Rho1, Rho2, Rho3, dx1, dx2, V, dt)

  performs one modified upwind step at constant velocity, V
  for time dt. For outflow boundary conditions. 
  
  Performs:

  d(RhoG)/dt = - V*d(Rho)/dx - (V/L)(RhoG - RhoAmb)

  where V is assumed positive (characteristic speed away from the
  grid), L is a characteristic length over which RhoG decays to
  RhoAmb. 

  An upwind discretization is used, and evaluated semi-implicitly. 
  Implicitness is important because V may be much larger than flow
  velocities in subsonic problems, thus causing (vdt/dr >> 1). 
  Thus...transition to extrapolated values is automatic.

  RhoN is the value just inside the computational grid. RhoG is the
  value just before the bc is applied. RhoAmb is the ambient value
  at distance L. dr is the drid spacing between guard cell and last
  computational cell. V is the characteristic speed (assumed positive).
  dt is the time step.

  Used only for outflow boundary conditions.

************************************************************/
static double MD_FCT_upwind_outflow(RhoG, RhoN, RhoAmb, dr, L, V, dt)
double RhoG, RhoN, RhoAmb, dr, L, V, dt;
{
double x, y, z, result;

x = (V*dt); y = L*dr; z = L+dr;

result = (y*RhoG + x*(L*RhoN + dr*RhoAmb))/(y + x*z);
return(result);


}
/************************************************************

  MD_chemstep(rlist, splist, dt, first)

  Evaluates and updates all reactions in rlist.  Each reaction is
  advanced the grid associated with the reaction list for
  all species in splist which are defined on that grid.

  Explicit first order evaluation is performed, but
  is converted to second order automatically when used
  with MD_Richardson. 

  dt is the time step.

  If first == true, takes rho_n to rho_np1. Otherwise,
  takes rho_np1 to rho_np1. 

  Makes assumptions about grids being all the same for
  all species. Covers all of space.

  returns maximum relative change, scaled to maximum value:

  (maximum increment)/(maximum value). 

  for use in time step control.

  Note: this is a time explicit method, and so is dangerous.
  Someday we'll have an asymptotic method.


************************************************************/
double MD_chemstep(rlist, splist, dt, first)
MD_reaction_list_ptr rlist;
MD_FCT_species_set_ptr splist;
double dt;
bool first;
{
  static MD_reaction_list_ptr old_rlist = NULL;
  MD_FCT_species_set_rec_ptr rec;
  MD_FCT_species_ptr sp;
  MD_reaction_ptr rptr;
  double max_frac, lr, r, d, maxv;
  unsigned int i, N, sl, ssteps, s, ir, sstart, istart, istop;
  static unsigned int old_version = 0, N_spe = 0;
  static dblptr max_incr=NULL, max_val=NULL;
  axis_ptr axs;

  if (rlist->size == 0) {return(0.0);}
  MD_current_step = MD_CHEMSTEP_STEP;
  grid = rlist->grid;
  MD_FCT_init_grid_data();
  N_DIM = grid->num_dim;
  /*
    initialize all pointers for reactions.
  */
  for(rptr = rlist->start; rptr != NULL; rptr = rptr->next){
    N = rptr->N_reactants;
    if (first) {
      for(i=0; i<N; i++){
	rptr->rho[i] = rptr->reactants[i]->rho_n;
      }
    } else {
      for(i=0; i<N; i++){
	rptr->rho[i] = rptr->reactants[i]->rho_np1;
      }
    }
  }

  if ((splist->number != N_spe)||(old_rlist != rlist)||(old_version != splist->
version)) {
    N_spe = splist->number;
    max_incr = get_memory(max_incr, N_spe, double);
    max_val = get_memory(max_val, N_spe, double);
    old_version = splist->version;
    old_rlist = rlist;
  }
  /*
    Go through the entire array of grid points, looking only at INTERIOR points
    (unless it's a zero-dimensional grid)
  */
  max_frac = 0.0;
  j = 0;
  for(i=0; i<N_spe; i++){
    max_incr[i] = max_val[i] = 0.0;
  }
  for(axs = grid->axes,j=0; ; j++){     /* looking at the j axis */
    sl = look_at_j_axis();
    do{
      if (axs != NULL) {
	ssteps = axs->mul;
	sl = scan_surface_j(sl);
	sstart = sl + ssteps; 
	istart = 1;
	istop = axs->num_pts;
      } else {
	ssteps = 0;
	sstart = 0;
	istart = 0;
	istop = 0;
      }
      for(s = sstart, i = istart; i <= istop; s += ssteps, i++){ /* go down 
each axis */
	grid->pos = s;
	/* 
	   zero all relevant species increments 
	   */
	for(rec = splist->start; rec != NULL; rec = rec->next){
	  sp = rec->first;
	  if (sp->grid == grid){
	    sp->increment = 0;
	  }
	}
	/*
	  Go through all reactions and update their last_rate
	  Also, update all species increments.
	  */
	for(rptr = rlist->start; rptr != NULL; rptr = rptr->next){
	  lr = rptr->last_rate = rptr->net_rate(rptr, grid);
	  for(ir=0; ir<rptr->N_reactants; ir++){
	    rptr->reactants[ir]->increment += lr*(rptr->delta)[ir]*dt;
	  }
	}
	/* 
	   Go through all species, and update their concentrations into rho_np1
        */
	for(rec = splist->start; rec != NULL; rec = rec->next){
	  sp = rec->first;
	  if (sp->grid == grid){
	    d = sp->increment;
	    if (first){
	      r = sp->rho_np1[s] = sp->rho_n[s] + d;
	    } else {
	      r = sp->rho_np1[s] += d;
	    }
	    max_val[sp->sequence] = max(ABS(r),(max_val[sp->sequence]));
	    max_incr[sp->sequence] = max(ABS(d), (max_incr[sp->sequence]));
	  }
	}
      }
    } while (NOT_DONE_WITH_j);
    if (axs == NULL) {break;} /* this allows zero dimensional grid to go 
through once */
    axs = axs->next;
    if (axs == NULL) {break;}
  }  
  for(rec = splist->start; rec != NULL; rec = rec->next){
    sp = rec->first;
    if (sp->grid == grid){
      i = sp->sequence;
      if ((maxv = max_val[i]) > 0.0){
	max_frac = max(max_frac, ((max_incr[i]/maxv)));
      }
    }
  }
  return(max_frac);
}






