/************************************************************

  MD_FCT_defs.h

  Created by Edward A. Richley (richley@mailaps.org)

  This file is subject to the terms and conditions of the 
  GNU General Public License. 

************************************************************/
#ifndef MD_FCT_defs_flag
#define MD_FCT_defs_flag
#include "basics.h"
#define MD_Diffusion_re_use_Dkappa 0x01
#define MD_Diffusion_first_fractional_step 0x08
#define ZERO 0.0
#define SYMMETRIC 1.0
#define ANTI_SYMMETRIC -1.0

extern FILE *MD_LOGFILE;

enum MD_STEP_TYPE {NONE, MD_FCT_STEP, MD_DIFFUSION_STEP, MD_CHEMSTEP_STEP};
typedef enum MD_STEP_TYPE MD_STEP;

enum MD_FCT_Runge_Kutta_step {half, full};
typedef enum MD_FCT_Runge_Kutta_step MD_FCT_flow_step;

enum MD_FCT_limiter_style {weak, strong, medium};
typedef enum MD_FCT_limiter_style MD_FCT_limiter;

enum boundary {left, right, local, periodic, remote, constant};
enum MD_FCT_convection_style {conservation, geometric_conservation, 
				flux_conservation, geometric_flux_conservation,
				advection, zip, compression, none};

typedef enum MD_FCT_convection_style MD_FCT_convection;

typedef MD_FCT_convection *MD_FCT_convection_ptr;

typedef struct MD_diffusivity{
  double D;
  double kappa;
} MD_diffusivity;

/*
  num_pts: number of internal points (0 and N+1 are added)
  index:   used for current position
  sequence:the position in the linked list
  mul:     factor used to find the absolute position in an array
*/
typedef struct axis{
  string name;
  string units;
  unsigned int num_pts;
  unsigned int index;
  unsigned int sequence;
  unsigned int mul;
  void (*dimfn)();
  double length;
  dblptr r;
  dblptr Area;
  dblptr Volume;
  struct axis *next;
  /*  void (*adjust)();*/
} axis;

typedef axis *axis_ptr;

/*
  Struct for each velocity of a species. 
  Each velocity is for a particular axis.
  Each velocity has a particular convection type.
  There can be any number of velocity fields for
  each axis.
*/
typedef struct MD_FCT_vel_list_rec{
  dblptr velocity;
  string name;
  string units;
  axis_ptr axis;
  MD_FCT_convection convection;
  struct MD_FCT_vel_list_rec *next;
} MD_FCT_vel_list_rec;

typedef MD_FCT_vel_list_rec *MD_FCT_vel_list_rec_ptr;

typedef struct geometry{
  string name;
  unsigned int alloc_size;
  unsigned int num_dim;
  unsigned int version;
  unsigned int pos;
  axis_ptr axes;
} geometry;

typedef geometry *geometry_ptr;

typedef struct MD_FCT_bc{
  double Sl;
  double Vl;
  bool constant_L;
  double rho_amb_l;
  double Lchar_l;
  double vchar_l;
  void (*Fl)();
  unsigned_int_ptr OFFSETl;
  double Sr;
  double Vr;
  bool constant_R;
  double rho_amb_r;
  double Lchar_r;
  double vchar_r;
  void (*Fr)();
  unsigned_int_ptr OFFSETr;
} MD_FCT_bc;

typedef MD_FCT_bc *MD_FCT_bc_ptr;

/*
  sequence: position in linked list
*/
typedef struct MD_FCT_species{
  string name;
  string units;
  geometry_ptr grid;
  MD_FCT_bc_ptr bc;
  unsigned int sequence;
  double mass;
  double charge;
  dblptr rho0;
  dblptr rhoS;
  dblptr rho1;
  dblptr rho;
  dblptr rhoS_init;
  dblptr rho_n;
  dblptr rho_np1;
  dblptr rho_mid;
  MD_FCT_vel_list_rec_ptr vel;
  dblptr *D;
  MD_diffusivity (*Dfn)();
  unsigned int figure;
  string FigureProperties;
  string ViewAngle;
  bool plot;
  bool Interactive_Plot;
  double increment;
  MD_FCT_limiter Flux_Limiter;
  bool enable_minimum;
  bool enable_maximum;
  double minimum;
  double maximum;
} MD_FCT_species;

typedef MD_FCT_species *MD_FCT_species_ptr;

typedef struct MD_FCT_species_set_rec{
  MD_FCT_species_ptr first;
  struct MD_FCT_species_set_rec *next;
} MD_FCT_species_set_rec;

typedef MD_FCT_species_set_rec *MD_FCT_species_set_rec_ptr;

typedef struct MD_FCT_species_set{
  string name;
  unsigned int number;
  unsigned int version;
  unsigned int focus;
  bool Initial_Plot;
  MD_FCT_species_set_rec_ptr start;
} MD_FCT_species_set;

typedef MD_FCT_species_set *MD_FCT_species_set_ptr;

/*

  MD_reaction:

  name: name of reaction
  N_reactants: number of participating reactants
  sequence: position in parent reaction list
  net_rate: function implementing rate evaluation 
            in the form: double net_rate(reaction, grid);
  last_rate: place to put the last evaluation of net_rate
  delta: vector of change quantities for each reactant (difference in 
      stoichiometric coefficients, or energy release, etc.)
  reactants: vector of poitners to reactant species.
  rho: vector of pointers to arrays of reactant concentrations
      from which to evaluate rate.
  kn: vector of parameters for reaction
  
*/
typedef struct MD_reaction{
  string name;
  unsigned int N_reactants;
  unsigned int sequence;
  double (*net_rate)();
  double last_rate;
  dblptr delta;
  MD_FCT_species_ptr *reactants;
  dblptr *rho;
  dblptr kn;
  struct MD_reaction *next;
} MD_reaction;

typedef MD_reaction *MD_reaction_ptr;

typedef struct MD_reaction_list{
  string name;
  unsigned int size;
  geometry_ptr grid;
  MD_reaction_ptr start;
  MD_FCT_species_set_ptr species;
} MD_reaction_list;

typedef MD_reaction_list *MD_reaction_list_ptr;

#endif






