/************************************************************

  MD_FCT_utils.c 

  Auxilliary routines used with MD_FCT for multi-dimensional 
  flow problems. See "MD_FCT_defs.h" for definitions of various 
  types, and be sure to include "MD_FCT.h" for function
  declarations.

  Created by Edward A. Richley (richley@mailaps.org)

  This file is subject to the terms and conditions of the 
  GNU General Public License. 


************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifdef ccc
#include <cpml.h>
#endif
#include "basics.h"
#include "MD_FCT_defs.h"
#include "MD_FCT.h"

#define MD_N_params 100

static void MD_FCT_get_data_space();
static MD_FCT_species_ptr MD_FCT_make_new_species();
static void MD_FCT_print_rho();
static void click_a_grid();
static unsigned int get_pos();
static void MD_dummy_bc_fn(MD_FCT_bc *);

extern FILE *MD_LOGFILE;
MD_FCT_flow_step MD_FCT_step;
FILE *MD_FCT_OUTPUT_FILE = NULL;
string MD_FCT_OUTPUT_FILE_NAME = "MD_FCT.out";
FILE *MD_FCT_INPUT_FILE = NULL;
string MD_FCT_INPUT_FILE_NAME = NULL;
bool MD_FCT_INPUT_FILE_FLAG = FALSE;
string MD_DATADIR = ".";

string MD_FCT_convection_names[8] = {"conservation", 
				       "geometric_conservation",
				       "flux_conservation",
				       "geometric_flux_conservation",
				       "advection", 
				       "zip",
 				       "compression",
				       "none"};

MD_FCT_species_ptr MD_get_spptr();
axis_ptr MD_get_axptr();
void MD_output_vel(MD_FCT_species_ptr, MD_FCT_vel_list_rec_ptr);
void MD_output_all_vel(MD_FCT_species_ptr);
unsigned int MD_get_position(geometry_ptr);
void MD_FCT_print_all_rho(MD_FCT_species_set_ptr);
void MD_output_all_rho(MD_FCT_species_set_ptr);
void MD_FCT_print_vel(MD_FCT_species_ptr);
void get_an_index(geometry_ptr, unsigned_int_ptr, unsigned_int_ptr, dblptr);
void MD_FCT_read_all_rho(MD_FCT_species_set_ptr);

static unsigned int one = 1;
/************************************************************

  MD_get_position()

  gets current position pointed to by all axes in "grid"

************************************************************/
unsigned int MD_get_position(grid)
geometry_ptr grid;
{
axis_ptr axs;
unsigned int result;
result = 0;
for(axs = (grid->axes); axs != NULL; axs = (axs->next)){
  result += axs->mul*(axs->index);
}
return(result);
}
/************************************************************

  MD_FCT_rect()

  function for dimensioning rectangular axes:

            0  ........................Length
            |                            |
Volume:   0 | 1 | 2 | .... | i | ... | N | N+1 |  
r:        0   1   2     i-1 i+1   N-1  N
Area:       0   1   2     i-1 i+1   N-1  N

equally spaced cells with size Length/N.

************************************************************/
void MD_FCT_rect(ax, grid, pos)
axis_ptr ax;
geometry_ptr grid;
unsigned int pos;
{
unsigned int i, N;
double dx;

N = ax->num_pts;
dx = ax->length/(double)N;
i = ax->index;

ax->r[pos] = ((double)i - .5)*dx;
ax->Area[pos] = 1.0;
ax->Volume[pos] = dx;
return;
}
/************************************************************

  MD_FCT_cyl()

  function for dimensioning cylindrical axes:

            0  ........................Length
            |                            |
Volume:   0 | 1 | 2 | .... | i | ... | N | N+1 |  
r:        0   1   2     i-1 i+1   N-1  N
Area:       0   1   2     i-1 i+1   N-1  N

equally spaced cells with size Length/N.

************************************************************/
void MD_FCT_cyl(ax, grid, pos)
axis_ptr ax;
geometry_ptr grid;
unsigned int pos;
{
unsigned int i, N;
double r, dr;

N = ax->num_pts;
dr = ax->length/(double)N;
i = ax->index;
r =  ((double)i)*dr;
ax->r[pos] = ((double)i-.5)*dr;
ax->Area[pos] = 2*pi*r;
ax->Volume[pos] = (pi*dr)*(2*r - dr);

return;
}
/************************************************************

  MD_FCT_sphere()

  function for dimensioning spherical axes:

            0  ........................Length
            |                            |
Volume:   0 | 1 | 2 | .... | i | ... | N | N+1 |  
r:        0   1   2     i-1 i+1   N-1  N
Area:       0   1   2     i-1 i+1   N-1  N

equally spaced cells with size Length/N.

************************************************************/
void MD_FCT_sphere(ax, grid, pos)
axis_ptr ax;
geometry_ptr grid;
unsigned int pos;
{
unsigned int i, N;
double r, dr;

N = ax->num_pts;
dr = ax->length/(double)N;
i = ax->index;
r = ((double)i)*dr;
ax->r[pos] = ((double)i-.5)*dr;
ax->Area[pos] = 4*pi*r*r;
ax->Volume[pos] = (4*pi*dr)*(r*r-r*dr+dr*dr/3.0);

return;
}

/************************************************************

  Axis definition routines for polar coordinates,
  with r and theta axes.

************************************************************/

#define DR_DEFAULT 1.0
#define Nr_DEFAULT 10
#define NT_DEFAULT 10
#define DT_DEFAULT pi/NT_DEFAULT

static double delta_theta = DT_DEFAULT, delta_r = DR_DEFAULT;
static unsigned_int_ptr r_index_ptr;
static axis_ptr r_axis, theta_axis;
static bool r_def = false, theta_def = false;

/************************************************************

  MD_polar_r()

  Axis initialization routine. 
  Dimensions radius axes for use with
  polar coordinates (r, theta).

            0  ........................Length
            |                            |
Volume:   0 | 1 | 2 | .... | i | ... | N | N+1 |  
r:          0   1   2     i-1 i+1   N-1  N
Area:       0   1   2     i-1 i+1   N-1  N

equally spaced cells with size Length/N.

  Makes new Area and Volume vectors, depending on 
  delta_theta.

************************************************************/
void MD_polar_r(ax, grid, pos)
axis_ptr ax;
geometry_ptr grid;
unsigned int pos;
{
unsigned int i, N;
double r, dr, Length, Vol;

Length = ax->length;
N = ax->num_pts;
delta_r = (dr = Length/(double)N);
i = ax->index;

Vol = .5*delta_theta*dr*dr;
if (theta_def) {
  delta_theta = theta_axis->length/theta_axis->num_pts;
}

r = (ax->r[pos] = ((double)i)*dr);
ax->Area[pos] = delta_theta*r;
ax->Volume[pos] = Vol*((double)(2*i)-1.0);

r_index_ptr = &(ax->index);
r_axis = ax;
r_def = true;
return;
}
/************************************************************

  MD_polar_theta()

  Axis initialization routine. 
  Dimensionins azimuth axis
  for use with polar coordinates (r, theta)

            0  ........................Aperture
            |                            |
Volume:   0 | 1 | 2 | .... | i | ... | N | N+1 |  
r:          0   1   2     i-1 i+1   N-1  N
Area:       0   1   2     i-1 i+1   N-1  N

equally spaced cells with size delta_theta = Aperture/N.  Area and
Volume depend on r position.  Makes Area and Volume vectors, depending
on r position, as given by r_index_ptr.



************************************************************/
void MD_polar_theta(ax, grid, pos)
axis_ptr ax;
geometry_ptr grid;
unsigned int pos;
{
unsigned int i,j,N;
double th, dth, Aperture, dr;
axis_ptr axs;

if (r_def) {
  j = (*r_index_ptr);
  delta_r = r_axis->length/r_axis->num_pts;
} else {
  j = Nr_DEFAULT;
}

N = ax->num_pts;
Aperture = ax->length;
delta_theta = (dth = Aperture/(double)N);
i = ax->index;
dr = delta_r;

th = (ax->r[pos] = ((double)i)*dth);
ax->Area[pos] = dr;
ax->Volume[pos] = (.5*delta_theta*dr*dr*((double)(2*j)-1.0));

theta_def = true;
theta_axis = ax;

return;
}
/************************************************************

  MD_FCT_get_data_space()

  Allocates memory for rho0, rhoS, and rho1
  of species sp.

  Initializes sp->rho0 and sp->vel to 0.0;

************************************************************/
static void MD_FCT_get_data_space(sp)
MD_FCT_species_ptr sp;
{
unsigned int size, n_dim, i;
size = sp->grid->alloc_size;
n_dim = max(1,sp->grid->num_dim);
 sp->rho0 = get_memory(sp->rho0, size, double);
 for(i=0; i<size; i++){
   sp->rho0[i] = 0.0;
 }
sp->rhoS = get_memory(sp->rhoS, size, double);
sp->rho1 = get_memory(sp->rho1, size, double);
sp->D = get_memory(sp->D, n_dim, dblptr);
sp->rho = sp->rho0;
sp->rhoS_init = sp->rho0;
if ((sp->rho0 == NULL)|(sp->rho1 == NULL)|(sp->bc == NULL)
    |(sp->rhoS == NULL)){
  fprintf(stderr,"ERROR: Insufficient memory [MD_FCT_get_data_space]\n");
  exit(1);
}
return;
}
/************************************************************

  MD_FCT_print_all_rho()

  Routine to print all concentration data for a species set.

************************************************************/
void MD_FCT_print_all_rho(set)
MD_FCT_species_set_ptr set;
{
MD_FCT_species_set_rec_ptr rec;
MD_FCT_species_ptr sp = NULL;

fprintf(MD_FCT_OUTPUT_FILE,"set: %s\n", set->name);
fprintf(MD_FCT_OUTPUT_FILE,"MD_Total_time: ");
fwrite((char *)&MD_Total_time, sizeof(double), 1, MD_FCT_OUTPUT_FILE);
fprintf(MD_FCT_OUTPUT_FILE,"\n");
for(rec = set->start; rec != NULL; rec = rec->next){
  sp = rec->first;
  MD_FCT_print_rho(sp);
}
return;
}
/************************************************************

  MD_output_all_rho()

  Routine to print all concentration data for a species set.
  Prints in format compatible with gnuplot.
  Prints into directory MD_DATADIR under file named
  species".dat". Old data is overwritten. 

************************************************************/
void MD_output_all_rho(set)
MD_FCT_species_set_ptr set;
{
MD_FCT_species_set_rec_ptr rec;
MD_FCT_species_ptr sp = NULL;
char flname[100];
int len;
FILE *outfile = NULL;
unsigned int i, pos, num_dim;
geometry_ptr grid;
axis_ptr ax;

for(rec = set->start; rec != NULL; rec = rec->next){
  sp = rec->first;

  strcpy(flname,MD_DATADIR);
  len = strlen(flname);
  if (flname[len] != '/') {
    strcat(flname,"/");
  }
  strcat(flname,sp->name);
  strcat(flname,".dat");
  outfile = fopen(flname,"w");

  fprintf(outfile,"# set: %s\n", set->name);
  grid = sp->grid;
  num_dim = grid->num_dim;
  fprintf(outfile,"# Species: %s Units: %s Grid: %s of dimension %d\n",
	  sp->name, sp->units, grid->name, num_dim);
  fprintf(outfile,"# MD_Total_time: %8.5e\n",MD_Total_time);
  fprintf(outfile,"# \n");

  for(ax = grid->axes; ax != NULL; ax = ax->next){ax->index = 0;}

  if (grid->num_dim > 0) {               /* make sure there is 1 or more axes */
    for(i=0; i<grid->alloc_size; i++){
      for(ax = grid->axes; ax != NULL; ax = ax->next){
	pos = MD_get_position(grid);
	fprintf(outfile,"%8.5e ",ax->r[pos]);
      }
      fprintf(outfile," %8.5e \n",sp->rho0[pos]);
      click_a_grid(grid);
      if (grid->axes != NULL) {
	if (grid->axes->index == 0) {fprintf(outfile,"\n");}
      }
    }
  } else {                              /* treat zero dimensional grid specially */
    fprintf(outfile,"0.0  %8.5e \n",sp->rho0[0]);
  }
  fprintf(outfile,"\n");
  fclose(outfile);
}
return;
}
/************************************************************

  MD_output_vel(sp, vel)

  Routine to print velocity data for a species. 
  
  Prints in format compatible with gnuplot.
  Prints into directory MD_DATADIR under file named
  sp->name.vel->name".dat". Old data is overwritten. 
  

************************************************************/
void MD_output_vel(sp, vel)
     MD_FCT_species_ptr sp;
     MD_FCT_vel_list_rec_ptr vel;
{
  MD_FCT_vel_list_rec_ptr ivel = NULL;
  char flname[100];
  int len;
  FILE *outfile = NULL;
  unsigned int i, pos, num_dim;
  geometry_ptr grid;
  axis_ptr ax;
  bool found;

  /* test to see if vel belongs to sp */
  found = false;
  for(ivel = sp->vel; ivel != NULL; ivel = ivel->next){
    if (vel == ivel){ found = true; break;}
  }
  
  if (found){
    strcpy(flname,MD_DATADIR);
    len = strlen(flname);
    if (flname[len] != '/') {
      strcat(flname,"/");
    }
    strcat(flname,sp->name);
    strcat(flname,".");
    strcat(flname,vel->name);
    strcat(flname,".dat");
    outfile = fopen(flname,"w");
  
    fprintf(outfile,"# species: %s\n", sp->name);
    grid = sp->grid;
    num_dim = grid->num_dim;
    fprintf(outfile,"# Velocity: %s Units: %s Grid: %s of dimension %d\n",
	    vel->name, vel->units, grid->name, num_dim);
    fprintf(outfile,"# MD_Total_time: %8.5e\n",MD_Total_time);
    fprintf(outfile,"# \n");
    
    for(ax = grid->axes; ax != NULL; ax = ax->next){ax->index = 0;}
    
    for(i=0; i<grid->alloc_size; i++){
      for(ax = grid->axes; ax != NULL; ax = ax->next){
	pos = MD_get_position(grid);
	fprintf(outfile,"%8.5e ",ax->r[pos]);
      }
      fprintf(outfile," %8.5e \n",vel->velocity[pos]);
      click_a_grid(grid);
      if (grid->axes != NULL) {
	if (grid->axes->index == 0) {fprintf(outfile,"\n");}
      }
    }
    fprintf(outfile,"\n");
    fclose(outfile);
  } else {
    fprintf(stderr,"[MD_output_vel]: Invalid velocity %s for species %s\n",
	    vel->name, sp->name);
  }	    
  return;
}
/************************************************************

  MD_output_all_vel()

  Routine to print all velocity data for a species. 
  Prints in format compatible with gnuplot.
  Prints into directory MD_DATADIR under file named
  species_name.velocity_name".dat". Old data is overwritten. 

************************************************************/
void MD_output_all_vel(sp)
     MD_FCT_species_ptr sp;
{
  MD_FCT_vel_list_rec_ptr ivel = NULL;
  char flname[100];
  int len;
  FILE *outfile = NULL;
  unsigned int i, pos, num_dim;
  geometry_ptr grid;
  axis_ptr ax;

  for(ivel = sp->vel; ivel != NULL; ivel = ivel->next){
     MD_output_vel(sp, ivel);
  }
  return;
}
/************************************************************

  MD_FCT_print_rho()

  Routine to print values of rho0. 
  Grid data is written as ASCII; rho data is binary.

************************************************************/
static void MD_FCT_print_rho(sp)
MD_FCT_species_ptr sp;
{
  axis_ptr axs;
  geometry_ptr grid;
  int i;

  grid = sp->grid;
  fprintf(MD_FCT_OUTPUT_FILE, 
	  "species: %s units: %s num_dim: %u\n",sp->name, sp->units, grid->num_dim);
  for(axs = (grid->axes); axs != NULL; axs = (axs->next)){
    fprintf(MD_FCT_OUTPUT_FILE, "axis: %s size: %u\n",axs->name, axs->num_pts);
  }
  fprintf(MD_FCT_OUTPUT_FILE, "rho: ");
  fwrite((void *)sp->rho0, sizeof(double), grid->alloc_size, MD_FCT_OUTPUT_FILE);
  fprintf(MD_FCT_OUTPUT_FILE,"\n;\n");
  return;
}
/************************************************************

  MD_FCT_print_vel()

  Routine to print values of velocity.
  Grid data is written as ASCII; velocity data is binary.
  Each velocity is printed separately, with titles giving
  axis name and velocity name.

************************************************************/
void MD_FCT_print_vel(sp)
MD_FCT_species_ptr sp;
{
  axis_ptr axs;
  geometry_ptr grid;
  MD_FCT_vel_list_rec_ptr ivel;

  grid = sp->grid;
  if (MD_FCT_OUTPUT_FILE == NULL){
    fprintf(stderr,"MD_FCT_OUTPUT_FILE undefined\n");
    return;
  }
  fprintf(MD_FCT_OUTPUT_FILE, "species: %s num_dim: %u\n",sp->name, grid->num_dim);
  for(axs = (grid->axes); axs != NULL; axs = (axs->next)){
    fprintf(MD_FCT_OUTPUT_FILE, "axis: %s size: %u\n",axs->name, axs->num_pts);
  }
  for(ivel = sp->vel; ivel != NULL; ivel = ivel->next){
    fprintf(MD_FCT_OUTPUT_FILE, "velocity \"%s\" on axis \"%s\":\n",
	    ivel->name, ivel->axis->name);
    fwrite((char *)ivel->velocity, sizeof(double), 
	   grid->alloc_size, MD_FCT_OUTPUT_FILE);
    fprintf(MD_FCT_OUTPUT_FILE,"\n;\n");
  }
  fprintf(MD_FCT_OUTPUT_FILE,"\n;\n");
  return;
}
/************************************************************

  click_a_grid(grid)

  increments axis indexes by one step

************************************************************/
static void click_a_grid(grid)
geometry_ptr grid;
{
axis_ptr axs;
unsigned int index;

for(axs = grid->axes; axs != NULL; axs = axs->next){
  index = axs->index+1;
  axs->index = index%(axs->num_pts+2);
  if(axs->index == index) {break;}
}
return;
}
/************************************************************
  
  get_pos(n_dim, mul, index)

  returns value of position pointed to by index

************************************************************/
static unsigned int get_pos(n_dim, mul, index)
unsigned int n_dim, *mul, *index;
{
unsigned int i, sum;

sum = 0;
for(i=0; i<n_dim; i++){
  sum += mul[i]*index[i];
}
return(sum);
}
/************************************************************
  
  get_an_index(grid, index, size, lambda)

  Finds index nearest to what is pointed to by grid.
  Also finds fractional portions along each axis for interpolation.

************************************************************/
void get_an_index(grid, index, size, lambda)
geometry_ptr grid;
unsigned int *index, *size;
dblptr lambda;
{
unsigned int i, num, den;
int ind;
double lam;
axis_ptr axs;

for(i=0, axs = grid->axes; axs != NULL; i++, axs = axs->next){
  num = (2*axs->index-1)*size[i]+axs->num_pts;
  den = 2*axs->num_pts;
  ind = num/den;
  lam = (double)(num - ind*den)/(double)den;
  if (lam >= .5) {
    lam = lam - 1.0;
    ind++;
  }
  if (ind < 0) {ind = 0; lam = 0.0;}
  if (ind > (size[i]+1)) {ind = size[i]+1; lam = 0.0;}
  index[i] = ind; lambda[i] = lam;
}
return;
}
/************************************************************
  
  MD_FCT_read_all_rho()

  Routine to read all concentration data for a species set
  from file MD_FCT_INPUT_FILE. Data are re-mapped onto
  grid with dimensions as set by grid data of species set.

  Compatible with MD_FCT_print_all_rho(). 

  Some errors in the input file are detected, as from
  corruption, or sizes being incorrect. 

  Set name and MD_Total_time are read first.

  Note: data for each species in MD_FCT_INPUT_FILE are
  read into corresponding arrays in set. If species name is
  not found in set, it is ignored. Likewise, if a species
  name is not included in MD_FCT_INPUT_FILE, no corresponding
  data is transferred to set. This type of error will not
  be detected.

************************************************************/
void MD_FCT_read_all_rho(set)
MD_FCT_species_set_ptr set;
{
MD_FCT_species_ptr sp = NULL;
axis_ptr ax;
geometry_ptr grid;
char spname[100], axname[100], setname[100], spunits[100], instring[200]; 
unsigned int num_dim, alloc_size, last_alloc_size = 0, N_rho, j, tmp;
static unsigned int N_read, *size = NULL, last_num_dim = 0, *map = NULL, *mul 
= NULL;
static unsigned int *index = NULL, id;
unsigned int i;
int offset;
static dblptr rho0 = NULL, lambda = NULL;
double sum, base, correction;
bool fail, init;
FILE *TMPFILE;
static string tmpstr = NULL;

fail = false;
init = false;
TMPFILE = fopen("poscol.tmp","w");
tmpstr = get_memory(tmpstr, 20, char);
/*
  find each set of data in the file.
*/
while (!init) {
  N_read = fscanf(MD_FCT_INPUT_FILE,"set: %s\n",setname);
  fail = (N_read != 1);
  if (fail) {break;}
  N_read = fread(tmpstr,sizeof(char), 15, MD_FCT_INPUT_FILE);
  tmpstr[15] = '\0';
  fail = ((N_read != 15)||(strcmp(tmpstr,"MD_Total_time: ") != 0));
  if (fail) {break;}
  N_read = fread(&MD_Total_time, sizeof(double), 1, MD_FCT_INPUT_FILE);
  fail = (N_read != 1);
  if (fail) {break;}
  fail = (fscanf(MD_FCT_INPUT_FILE,"\n") == EOF);
  init = true;
}
set->name = strdup(setname);

 while ((fgets(instring, 200, MD_FCT_INPUT_FILE) != NULL)&& !fail){
   N_read = sscanf(instring,"species: %s units: %s num_dim: %u\n",spname,spunits,&num_dim);
   if (N_read != 3) {
        N_read = sscanf(instring,"species: %s units: num_dim: %u\n",spname,&num_dim);
	strcpy(spunits,"");
	fail = (N_read != 2);
   }

  if (fail) {break;}
  if (num_dim > last_num_dim){
    size = get_memory(size, (4*num_dim+1), unsigned int);
    map = size + num_dim;
    mul = map + num_dim;
    index = mul + num_dim + 1;
    lambda = get_memory(lambda, num_dim, double);
    last_num_dim = num_dim;
  }
  sp = MD_get_spptr(spname, set);
  if (sp != NULL){
    sp->units = strdup(spunits);
    alloc_size = 1;
    grid = sp->grid;
    mul[0] = 1;
    for(i=0; i<num_dim; i++){
      N_read = fscanf(MD_FCT_INPUT_FILE,"axis: %s size: %u\n",axname,&(size[i]));
      fail = (N_read != 2);
      if (fail) { break;}
      mul[i+1] = (size[i]+2)*mul[i];
      map[i] = (MD_get_axptr(axname, grid))->sequence;
    }
    alloc_size = mul[num_dim];
/*
  un-permute the axis data
*/
    for(i=0; i<num_dim; i++){
      for(j=0; (map[j] != i)&&(j < num_dim); j++){}
      if (j >= num_dim) {fail = true; break;}
      tmp = size[i]; size[i] = size[j]; size[j] = tmp;
      tmp = mul[i]; mul[i] = mul[j]; mul[j] = tmp;
      map[i] = i;
    }
    for(i = 0, ax = (grid->axes); ax != NULL; i++, ax = (ax->next)){
      ax->index = 0;
    }
  }
  if (fail) {break;}
  N_rho = fread((void *)tmpstr, sizeof(char), 5, MD_FCT_INPUT_FILE);
  tmpstr[5] = '\0';
  fail = ((N_rho != 5)||(strcmp(tmpstr,"rho: ") != 0));
  /*  fail = (fscanf(MD_FCT_INPUT_FILE,"rho: ")==EOF);*/
  if (fail) {break;}
  if (alloc_size > last_alloc_size){
    free(rho0);
    rho0 = get_memory(rho0, alloc_size, double);
    last_alloc_size = alloc_size;
  }
  N_rho = fread((void *)rho0, sizeof(double), alloc_size, MD_FCT_INPUT_FILE);
  fail = (N_rho != alloc_size);
  if (fail) {break;}
  fail = (fscanf(MD_FCT_INPUT_FILE,"\n;\n") == EOF);
  if (fail)  {break;}
/*
  pack the data into the array at sp->rho0. Re-map as necessary.
*/
  for(i=0; i<(grid->alloc_size); i++){
    get_an_index(grid, index, size, lambda);
    id = get_pos(num_dim, mul, index);
    sum = base = rho0[id];
    for(j=0; j<num_dim; j++){
      offset = ((lambda[j]>0.0)?mul[j]:-mul[j]);
      offset = ((offset > 0)&&(index[j]==size[j]))?0:offset;
      offset = ((offset < 0)&&(index[j] == 0))?0:offset;
      correction = ABS(lambda[j])*(rho0[id+offset]-base);
      sum += correction;      
    }
    id = MD_get_position(grid);
    sp->rho0[id] = sum;
    click_a_grid(grid);
  }
 }
if (fail) {
  fprintf(stderr,"MD_FCT_read_all_rho: error in input file\n");
  exit(1);
}
free(rho0);
free(size);
fclose(TMPFILE);
return;
}
/************************************************************

  MD_FCT_make_new_species()

  Allocates space for a new species named SPNAME on grid GRID.
  Initializes pointers and calls MD_FCT_get_data_space.

  bc parameters are all set to zero. OFFSETs are set to local
  condtions, and boundary condition function is set to 
  MD_dummy_bc_fn().

  Sets mass and charge to 0.0.

************************************************************/
static MD_FCT_species_ptr MD_FCT_make_new_species(SPNAME,UNITS, GRID)
string SPNAME, UNITS;
geometry *GRID;
{
MD_FCT_species_ptr spptr = NULL;
unsigned int i, n_dim;

spptr = NULL;
spptr = get_memory(spptr, 1, MD_FCT_species);
spptr->name = strdup(SPNAME);
spptr->units = strdup(UNITS);
spptr->grid = GRID;
n_dim = max(1,spptr->grid->num_dim);
spptr->bc = NULL;
spptr->bc = get_memory(spptr->bc, n_dim, MD_FCT_bc);
spptr->charge = 0.0;
spptr->mass = 0.0;
spptr->rho0 = spptr->rhoS = spptr->rho1 = NULL;
spptr->vel = NULL;
spptr->D = NULL;
spptr->rho_n = spptr->rho_np1 = spptr->rho_mid = NULL;
for(i=0; i<n_dim; i++){
  spptr->bc[i].Fl = spptr->bc[i].Fr = MD_dummy_bc_fn;
  spptr->bc[i].Sl = spptr->bc[i].Vl = spptr->bc[i].rho_amb_l = 
    spptr->bc[i].Lchar_l = spptr->bc[i].vchar_l = 0.0;
  spptr->bc[i].Sr = spptr->bc[i].Vr = spptr->bc[i].rho_amb_r = 
    spptr->bc[i].Lchar_r = spptr->bc[i].vchar_r = 0.0;
  spptr->bc[i].constant_L = spptr->bc[i].constant_R = false;
  (spptr->bc)[i].OFFSETl =  (spptr->bc)[i].OFFSETr = &(one);
}
spptr->figure = 0;
spptr->FigureProperties = "";
spptr->ViewAngle = "";
spptr->plot = true;
spptr->Interactive_Plot = true;
spptr->Flux_Limiter = strong;
spptr->enable_maximum = FALSE;
spptr->enable_minimum = FALSE;
spptr->maximum = 0.0;
spptr->minimum = 0.0;
spptr->sequence = 0;
MD_FCT_get_data_space(spptr);
for(i=0; i<n_dim; i++){
  spptr->D[i] = NULL;
}
return(spptr);
}
/************************************************************

  MD_FCT_create_velocity

  Creates velocity for species "spe" on axis "ax" with
  convection type "convection" name "name" and units "units".

  Returns pointer to the newly created velocity (if successful).

************************************************************/
MD_FCT_vel_list_rec_ptr MD_FCT_create_velocity(spe, ax, convection, name, units)
     string ax, name, units;
     MD_FCT_convection convection;
     MD_FCT_species_ptr spe;
{
  geometry_ptr grid;
  dblptr vel_array;
  axis_ptr axs;
  MD_FCT_vel_list_rec_ptr tvel, *ivel;
  unsigned int i, size;

  if (spe == NULL){
    fprintf(stderr,"Invalid species, can't create velocity \n");
    return(NULL);
  }
  grid = spe->grid;
  if ((axs = MD_get_axptr(ax, grid)) == NULL){
    fprintf(stderr,"[MD_FCT_create_velocity]: can't find axis %s for species %s\n",ax,spe->name);
    exit(1);
  }
  tvel = NULL;
  tvel = get_memory(tvel, 1, MD_FCT_vel_list_rec); 
  
  size = grid->alloc_size; 
  vel_array = NULL;
  vel_array = get_memory(vel_array, size, double);
  for(i=0; i<size; i++){
    vel_array[i] = 0.0;
  }
  tvel->velocity = vel_array;
  tvel->name = strdup(name);
  tvel->units = strdup(units);
  tvel->axis = axs; 
  tvel->convection = convection;
  for(ivel = &(spe->vel); *ivel != NULL; ivel = &((*ivel)->next)){}
  *ivel = tvel;
  return(tvel);
}
/************************************************************

  MD_FCT_bc_set()

  Sets "edge" (left or right) boundary conditions of
  species "spptr" for axis "axptr" on grid "grid" to S, V.
  Thus, for (edge == left) spe->bc[i].Sl = S and
  spe->bc[i].Vl = V where i is the sequence of axis "axptr"
  and "spptr" and "axptr" are given by pointer.  

  Note: If "periodic" boundary conditions are specified,
  does NOT assume S=1.0 and V=0.0. Calling routing
  must specify these as they are passed on to the bc.

************************************************************/
void MD_FCT_bc_set(spptr, axptr, edge, S, V, location)
axis_ptr axptr;
enum boundary edge, location;
double S, V;
MD_FCT_species_ptr spptr;
{
unsigned int i, *offset_ptr;
double slope, intercept;
bool conflag;

 slope = S;
 intercept = V;

conflag = false;
switch (location) {
 case local: 
  offset_ptr = &(one);
  break;
 case periodic: 
  offset_ptr = &(axptr->num_pts);
   break;
case constant:
  offset_ptr = &(one);
   conflag = true;
  break;
 default: offset_ptr = &(one);
  break;
}
i = axptr->sequence;
if (edge == left){
  (spptr->bc)[i].OFFSETl = offset_ptr;
  (spptr->bc)[i].Sl = slope;
  (spptr->bc)[i].Vl = intercept;
  (spptr->bc)[i].Lchar_l = 0.0;
  (spptr->bc)[i].vchar_l = 0.0;
  (spptr->bc)[i].constant_L = conflag;
  (spptr->bc)[i].Fl = NULL;
} else if (edge == right) {
  (spptr->bc)[i].OFFSETr = offset_ptr;
  (spptr->bc)[i].Sr = slope;
  (spptr->bc)[i].Vr = intercept;
  (spptr->bc)[i].Lchar_r = 0.0;
  (spptr->bc)[i].vchar_r = 0.0;
  (spptr->bc)[i].constant_R = conflag;
  (spptr->bc)[i].Fr = NULL;
} else {
  fprintf(stderr,"Invalid edge, can't set bc for %s\n",spptr->name);
}
return;
}
/************************************************************

  MD_FCT_set_bc()

  Sets "edge" (left or right) boundary conditions of
  species "spe" for axis "ax" on grid "grid" to S, V.
  Thus, for (edge == left) spe->bc[i].Sl = S and
  spe->bc[i].Vl = V where i is the sequence of axis "ax"
  and "ax" is given by string name.

************************************************************/
void MD_FCT_set_bc(spe, ax, edge, S, V, location)
string ax;
enum boundary edge, location;
double S, V;
MD_FCT_species_ptr spe;
{
geometry_ptr grid;
axis_ptr axs;

if (spe == NULL){
  fprintf(stderr,"Invalid species, can't set bc\n");
  return;
}
grid = spe->grid;

if ((axs = MD_get_axptr(ax, grid)) == NULL){
  fprintf(stderr,"[MD_FCT_set_bc]: can't find axis %s for species %s\n",ax,spe->name);
  exit(1);
}
MD_FCT_bc_set(spe, axs, edge, S, V, location);
return;
}
/************************************************************

  MD_FCT_set_bc_fn()

  Sets boundary condition at "edge" of species "spe" for
  axis "ax" to be: fn

  If (fn == NULL), no change is made. This allows
  the default function "MD_dummy_bc_fn()" to be kept.
  
  If function (fn != NULL), it is assumed that the
  provided function will affect S, V, L_char, and v_char
  so as to be useful. As so, these are unchanged.

  If it hasn't already been set, OFFSET is set to correspond 
  to local conditions as a default. 

  No other parameters are affected (location, S, V,
  L_char, v) so these must be specified
  by a call to MD_FCT_set_bc(), or directly modified
  outside fn().

************************************************************/
void MD_FCT_set_bc_fn(spe, ax, edge, fn)
MD_FCT_species_ptr spe;
string ax;
enum boundary edge;
void (*fn)();
{
  axis_ptr axs;
  unsigned int i, *offset_ptr;
  static unsigned int one = 1;
  enum boundary loc;
  geometry_ptr grid;
  double L, v, slope, intercept;
  
  if (spe == NULL){
    fprintf(stderr,"Invalid species, can't set bc\n");
    return;
  }
  grid = spe->grid;
  if ((axs = MD_get_axptr(ax, grid)) == NULL){
    fprintf(stderr,"[MD_FCT_set_bc_fn]: can't find axis %s for species %s\n",ax,spe->name);
    exit(1);
  }
  if (fn == NULL){return;}
  i = axs->sequence;

  offset_ptr = &(one);

  if (edge == left){
    if ((spe->bc)[i].OFFSETl == NULL) {
      (spe->bc)[i].OFFSETl = &(one);
    } 
    (spe->bc)[i].Fl = fn;
  } else if (edge == right){
    if ((spe->bc)[i].OFFSETr == NULL) {
      (spe->bc)[i].OFFSETr = &(one);
    }
    (spe->bc)[i].Fr = fn;
  } else {
    fprintf(stderr,"Invalid edge, can't set bc for %s\n",spe->name);
  }
return;
}
/************************************************************

  MD_FCT_new_species_set()

  Creates a new species set. Start is a pointer
  to the linked list of species.

************************************************************/
MD_FCT_species_set_ptr MD_FCT_new_species_set(name)
string name;
{
MD_FCT_species_set_ptr result = NULL;
result = get_memory(result, 1, MD_FCT_species_set);
result->name = strdup(name);
result->number = 0;
result->version = 0;
result->Initial_Plot = true;
result->start = NULL;
return(result);
}
    
/************************************************************

  MD_FCT_new_species()

  Generates an entry in the linked list beginning with
  set->species for a new species. Replaces existing
  species of the same name, or extends the list if
  name is new.

************************************************************/
MD_FCT_species_ptr MD_FCT_new_species(NAME, UNITS, set, grid)
string NAME, UNITS;
MD_FCT_species_set_ptr set;
geometry_ptr grid;
{
MD_FCT_species_set_rec_ptr new_rec_ptr = NULL, *rec;
MD_FCT_species_ptr spptr = NULL;
unsigned int i, index;

for(rec = &(set->start), index = 0; (*rec) != NULL; rec = &((*rec)->next), 
index++){
  if (!strcmp((*rec)->first->name,NAME)) {
    fprintf(stderr,"Replacing ");
    spptr = (*rec)->first;
    for(i=0; i<spptr->grid->num_dim; i++){
      if (spptr->D[i] != NULL) { free(spptr->D[i]); }
    }
    spptr->units = strdup(UNITS);
    free(spptr->D);
    spptr->D = NULL;
    spptr->grid = grid;
    set->version++;
    spptr->rho_mid = NULL;
    MD_FCT_get_data_space(spptr);
    break;
  }
}
if ((*rec) == NULL) {
  fprintf(stderr,"Initializing: ");
  new_rec_ptr = NULL;
  new_rec_ptr = get_memory(new_rec_ptr, 1, MD_FCT_species_set_rec);
  spptr = MD_FCT_make_new_species(NAME, UNITS, grid);
  new_rec_ptr->first = spptr;
  new_rec_ptr->next = NULL;
  spptr->sequence = index;
  set->version++;
  set->number++;
  *rec = new_rec_ptr;
}

fprintf(stderr,"species: %s on grid: %s\n",NAME,grid->name);
return(spptr);
}

/************************************************************

  MD_FCT_new_grid(name)

  Allocate space for a new grid named grid. 
  Begins linked list to set of axes.

************************************************************/
geometry *MD_FCT_new_grid(name)
string name;
{
geometry *result = NULL;
result = get_memory(result, 1, geometry);
result->name = strdup(name);
result->alloc_size = 1;
result->num_dim = 0;
result->version = 0;
result->axes = NULL;
return(result);
}


/************************************************************

  MD_FCT_new_axis(NAME, UNITS, N, grid, fn)

  generate an entry in grid->axes for a new axis with name NAME,
  and N number of interior points (plus two end points).
  UNITS is a string representing the physical units of the axis.
  Memory is allocated for axis dimensions r, Area, Volume.
  Replaces axis if it already exists, extends the
  list if it's a new one. Returns pointer to axis.

************************************************************/
axis_ptr MD_FCT_new_axis(NAME, UNITS, N,grid, Length, fn)
string NAME, UNITS;
unsigned int N;
geometry *grid;
double Length;
void (*fn)();
{
unsigned int num_dim, i, mul, alloc_size, pos;
int remainder;
static axis_ptr tmp_axis = NULL;
static unsigned_int_ptr muls = NULL;
static unsigned_int_ptr *indices = NULL;
axis_ptr *ax, axs;
bool already_exists, fail;
void (*dimfn)();
div_t position;

num_dim = grid->num_dim;
grid->version++;
tmp_axis = NULL;
tmp_axis = get_memory(tmp_axis, 1, axis);

already_exists = false;
mul = 1;
for(ax=&(grid->axes), i = 0; 
    ((*ax) != NULL)?
    !(already_exists = !strcmp(NAME, (*ax)->name)):false;
    ax = &((*ax)->next), i++){
  mul *= ((*ax)->num_pts+2);
}
if (!already_exists){
  tmp_axis->name = strdup(NAME);
  tmp_axis->units = strdup(UNITS);
  tmp_axis->num_pts = N;
  tmp_axis->index = 0;
  tmp_axis->sequence = i;
  tmp_axis->mul = mul;
  tmp_axis->dimfn = fn;
  tmp_axis->length = Length;
  tmp_axis->r = NULL;
  tmp_axis->r = get_memory(tmp_axis->r, (N+2), double);
  tmp_axis->Area = NULL;
  tmp_axis->Area = get_memory(tmp_axis->Area, (N+2), double);
  tmp_axis->Volume = NULL;
  tmp_axis->Volume = get_memory(tmp_axis->Volume, (N+2), double);
  tmp_axis->next = NULL;
  *ax = tmp_axis;
  grid->num_dim = num_dim+1;
  grid->alloc_size *= (N+2);
  fprintf(stderr,"Initializing: ");
} else {
  axs = *ax;
  grid->alloc_size = (grid->alloc_size/(axs->num_pts+2))*(N+2);
  axs->num_pts = N;
  axs->index = 0;
  axs->sequence = i;
  axs->mul = mul;
  axs->dimfn = fn;
  axs->length = Length;
  axs->r = get_memory(axs->r, (N+2), double);
  axs->Area = get_memory(axs->Area, (N+2), double);
  axs->Volume = get_memory(axs->Volume, (N+2), double);
  tmp_axis = axs;
  fprintf(stderr,"Replacing ");
  mul *= (axs->num_pts+2);
  for(ax=&axs; ((*ax) != NULL); ax = &((*ax)->next)){
    (*ax)->mul = mul;
    mul *= ((*ax)->num_pts+2);
  }
}
/**********

  fix up allocation of all of the axis arrays in the grid

**********/
alloc_size = grid->alloc_size;
fail = false;
num_dim = grid->num_dim;
muls = get_memory(muls, num_dim, unsigned int);
indices = get_memory(indices, num_dim, unsigned_int_ptr);
for(ax=&(grid->axes), i=0; (*ax != NULL); ax = &((*ax)->next), i++){
  (*ax)->r = get_memory((*ax)->r, alloc_size, double);
  (*ax)->Area = get_memory((*ax)->Area, alloc_size, double);
  (*ax)->Volume = get_memory((*ax)->Volume, alloc_size, double);
  muls[i] = (*ax)->mul;
  indices[i] = &((*ax)->index);
  dimfn = (*ax)->dimfn;
  if (dimfn == NULL){
    fprintf(stderr,"WARNING: No function given for dimensions of axis \"%s\"\n",
	    NAME);
    fail = true;
  }
}
if (fail) {
  fprintf(stderr,"Error: Can not continue\n");
  exit(1);
}
/******************

  step through all the positions in all axes and set up the 
  axis parameters. Get indices of each axis from position
  in array (pos).

*****************/
for(pos=0; pos<grid->alloc_size; pos++){
  remainder = (int)pos;
  for(i=num_dim; i!=0; i--){
    position = div(remainder,(int)muls[i-1]);
    remainder = position.rem;
    *(indices[i-1]) = position.quot;
  }
  for(ax = &(grid->axes); (*ax != NULL); ax = &((*ax)->next)){
    dimfn = (*ax)->dimfn;
    dimfn((*ax), grid, pos);
  }
}

fprintf(stderr,"axis %s with %u pts\n",NAME, N);
return(tmp_axis);
}

/************************************************************

  MD_get_spptr(NAME, SET)

  Gets pointer to species with name NAME from set SET.
  Returns NULL if not found.

************************************************************/
MD_FCT_species_ptr MD_get_spptr(NAME, SET)
string NAME;
MD_FCT_species_set_ptr SET;
{
MD_FCT_species_set_rec_ptr *rec;
MD_FCT_species_ptr spptr = NULL;

 if (NAME == NULL){ return(NULL);}
for(rec = &(SET->start); (*rec) != NULL; rec = &((*rec)->next)){
  if (!strcmp((*rec)->first->name,NAME)) {
    spptr = (*rec)->first;
    break;
  }
}
if (spptr == NULL){ 
  fprintf(stderr,"[MD_get_spptr]: Can't find species \"%s\"\n",NAME);
}
return(spptr);
}
/************************************************************

  MD_get_axptr()

  returns pointer to axis NAME in grid. 

************************************************************/
axis_ptr MD_get_axptr(NAME, grid)
geometry_ptr grid;
string NAME;
{
axis_ptr axs;
unsigned int i;

for(axs = (grid->axes), i=0; axs != NULL; axs = (axs->next),i++){
    if (!strcmp(axs->name, NAME)){
      break;
    }
  }
if(axs == NULL){
  fprintf(stderr,"[MD_get_axptr]: Can't find axis \"%s\"\n",NAME);
}
return(axs);
}
/************************************************************

  MD_new_reaction_list(name)

  Allocate space for a new reaction list named name.
  Begins linked list to set of reactions.

************************************************************/
MD_reaction_list_ptr MD_new_reaction_list(name, set)
string name;
MD_FCT_species_set_ptr set;
{
  MD_reaction_list_ptr result = NULL;
  result = get_memory(result, 1, MD_reaction_list);
  result->name = strdup(name);
  result->species = set;
  result->size = 0;
  result->start = NULL;
  result->grid = NULL;
  return(result);
}

/************************************************************

  MD_new_reaction(NAME, rlist, set, splist, fn)

  generate an entry in rlist->start for a new reaction 
  with name NAME, and  rate function fn. 
  set is the species set in which reactants will be found.
  splist is a NULL-terminated list of names  of reactants.
  deltas is a vector of net changes for each reactant in splist.
  ratefn(reaction,set) = pointer to function defining net reaction rate.
  kn is a pointer to a vector of parameters to pass to the reaction.
  Returns pointer to reaction.

************************************************************/
MD_reaction_ptr MD_new_reaction(NAME, rlist, set, splist, deltas, kn, ratefn)
string NAME;
MD_reaction_list_ptr rlist;
MD_FCT_species_set_ptr set;
string *splist;
dblptr deltas, kn;
double (*ratefn)();
{
unsigned int i, Nspe; 
static MD_reaction_ptr tmp_reaction = NULL;
MD_reaction_ptr this_reaction = NULL;
MD_reaction_ptr *reaction, react;
bool already_exists, fail;
div_t position;

tmp_reaction = NULL;
tmp_reaction = get_memory(tmp_reaction, 1, MD_reaction);

already_exists = false;
/* count the reactants */
for(Nspe=1; splist[Nspe] != NULL; Nspe++){}
/* see if this one already exists */
for(reaction=&(rlist->start), i = 0; 
    (*reaction != NULL)?
    !(already_exists = !strcmp(NAME, (*reaction)->name)):false;
    reaction = &((*reaction)->next), i++){
}
if (!already_exists){
  tmp_reaction->name = strdup(NAME);
  tmp_reaction->N_reactants = Nspe;
  tmp_reaction->sequence = i;
  tmp_reaction->net_rate = ratefn;
  tmp_reaction->delta = NULL;
  tmp_reaction->delta = get_memory(tmp_reaction->delta, Nspe, double);
  tmp_reaction->reactants = NULL;
  tmp_reaction->reactants = 
    get_memory(tmp_reaction->reactants, Nspe, MD_FCT_species_ptr);
  tmp_reaction->rho = NULL;
  tmp_reaction->rho = get_memory(tmp_reaction->rho, Nspe, dblptr);
  tmp_reaction->kn = kn;
  for(i=0; i<Nspe; i++){
    tmp_reaction->delta[i] = deltas[i];
    tmp_reaction->reactants[i] = MD_get_spptr(splist[i],set);
    fprintf(stderr,"%s\n",tmp_reaction->reactants[i]->name);
    if (tmp_reaction->reactants[i] == NULL){
      fprintf(stderr,"Unable to find species %s in set %s needed for reaction %s\n",
	      splist[i],set->name, tmp_reaction->name);
      exit(1);
    }
    tmp_reaction->rho[i] = tmp_reaction->reactants[i]->rho0;
  }
  tmp_reaction->next = NULL;
  *reaction = tmp_reaction; /* add new reaction to end of linked list */
  rlist->size++;

  fprintf(stderr,"Initializing: ");
} else {
  react = *reaction;
  react->N_reactants = Nspe;
  react->name = strdup(NAME);
  react->sequence = i;
  react->net_rate = ratefn;
  react->delta = get_memory(react->delta, Nspe, double);
  react->reactants = 
    get_memory(react->reactants, Nspe, MD_FCT_species_ptr);
  react->rho = get_memory(react->rho, Nspe, dblptr);
  react->kn= kn;
  for(i=0; i<Nspe; i++){
    react->delta[i] = deltas[i];
    react->reactants[i] = MD_get_spptr(splist[i],set);
    if (react->reactants[i] == NULL){
      fprintf(stderr,"Unable to find species \"%s\" in set \"%s\" needed for reaction \"%s\"\n",
	      splist[i],set->name, react->name);
      exit(1);
    }
    react->rho[i] = react->reactants[i]->rho0;
  }
  
  fprintf(stderr,"Replacing ");
}
fprintf(stderr,"reaction %s with %u reactants\n",NAME, Nspe);
return(tmp_reaction);
}

/************************************************************
  
  MD_integral_dV()

  Integrates vector y over "active" region of axis
  ax (1 thru num_pts) dVolume. Simple dot product
  to be consistent with finite difference representation.

  y[1] to y[num_pts] contain data points.
  ax is the axis pointer over which to integrate.

  y is a vector of length (num_pts+2).

************************************************************/
double MD_integral_dV(y, ax, grid)
dblptr y;
axis_ptr ax;
geometry_ptr grid;
{
unsigned int i, N, pos, ssteps; 
double sum;
dblptr dV;

N = ax->num_pts;
dV = ax->Volume;
sum = 0.0;
ssteps = ax->mul;
ax->index = 1;
pos = MD_get_position(grid);
for(i=1; i<=N; pos += ssteps, i++){
  sum += y[i]*dV[pos];
}
return(sum);
}
/************************************************************
  
  MD_integral_dVol()

  Integrates array y over "active" region of axis
  ax (1 thru num_pts) dVolume. Simple dot product
  to be consistent with finite difference representation.

  ax is the axis pointer over which to integrate.
  all other axis indexes are assumed to be in the right places.

  y is an array of size grid->alloc_size.

************************************************************/
double MD_integral_dVol(y, ax, grid)
dblptr y;
axis_ptr ax;
geometry_ptr grid;
{
unsigned int i, N, pos, ssteps; 
double sum;
dblptr dV;

N = ax->num_pts;
dV = ax->Volume;
sum = 0.0;
ssteps = ax->mul;
ax->index = 1;
pos = MD_get_position(grid);
for(i=1; i<=N; pos += ssteps, i++){
  sum += y[pos]*dV[pos];
}
return(sum);
}
/************************************************************
  
  MD_integral_dr()

  Integrates vector y over "active" region of axis
  ax (halfway between r[0] and r[1] thru halfway
  between r[num_pts] and r[num_pts+1]) dr. 
  Implements Hermite Cubic Quadrature algorithm
  as described by Kahaner, Moler, and Nash, chs. 4
  and 5. End points have strange coefficients because
  integration is not from ax->r[0] to ax->r[num_pts+1]. 

************************************************************/
double MD_integral_dr(y, ax, grid)
dblptr y;
axis_ptr ax;
geometry_ptr grid;
{
unsigned int i, p, N, ssteps, pos, last_pos;
double sum, hi, last_hi, deli, last_deli, alpha, beta, di, span;
dblptr r;

N = ax->num_pts;
r = ax->r;
ssteps = ax->mul;
ax->index = 1;
pos = MD_get_position(grid);
hi = r[pos] - r[pos-ssteps];
deli = (y[1] - y[0])/hi;
for(i=1, p=2; i<=N; i++, p++){
  last_pos = pos;
  pos += ssteps;
  last_hi = hi;
  last_deli = deli;
  hi = r[pos] - r[last_pos];
  span = hi+last_hi;
  deli = (y[p]-y[i])/hi;
  if (i == 1){
    di = (-last_hi*deli+last_deli*(last_hi+span))/span;
    alpha = THREE_THIRTYTWO*last_hi;
    beta = FIVE_ONENINETYTWO*last_hi*last_hi;
    sum = alpha*y[0]+beta*di;
  }
  di = (deli*last_hi + last_deli*hi)/span;
  alpha = .5*span;
  beta = alpha*(hi - last_hi)*ONE_SIXTH;
  sum += alpha*y[i] + beta*di;
}
di = (deli*(hi + span) - last_deli*hi)/span;
alpha = THREE_THIRTYTWO*hi;
beta = -FIVE_ONENINETYTWO*hi*hi;
sum += alpha*y[N]+beta*di;
return(sum);
}
/************************************************************

  MD_dY_dx()

  general purpose routine to differentiate vector Y
  with respect to vector x.

************************************************************/
double MD_dY_dx(Y, x, N, i)
dblptr Y, x;
unsigned int N, i;
{
double result, hi, hi_last, span, deli_last, deli;
unsigned int j;
bool low_lim, hi_lim;

low_lim = (i == 0); hi_lim = (i == N);
j = low_lim?1:(hi_lim?(N-1):i);
hi = x[j+1] - x[j];
hi_last = x[j] - x[j-1];
span = hi+hi_last;
deli = (Y[j+1] - Y[j])/hi;
deli_last = (Y[j] - Y[j-1])/hi_last;
result = (deli*hi_last + deli_last*hi);
if (low_lim||hi_lim){
  result += low_lim?(2.0*hi_last*(deli_last - deli)):
  (2*hi*(deli - deli_last));
}					    
result /= span;
return(result);
}  
/************************************************************

  MD_dy_dx()

  gets derivative estimate for vector "y"
  with respect to axis "ax" at position 
  set by ax->index.

  computation is done by quadratic interpolation among
  neighboring points. End points are treated similarly,
  but with two neighbors on one side. 

  Bad idea. Should be removed.


************************************************************/
/*double MD_dy_dx(y, ax, grid)
dblptr y;
axis_ptr ax;
{
double result;
unsigned int i, lim;
dblptr r;

r = ax->r;
i = ax->index;
lim = ax->num_pts+1;
result = MD_dY_dx(y,r,lim,i);

return(result);
}*/
/************************************************************

  MD_d_dr()

  gets derivative estimate for multi-dimensional array "y"
  with respect to axis "ax" at position "pos".
  "pos" is probably derived from a call to MD_get_position.
  The current index of "ax" must correspond to "pos". 

  Computation is done by quadratic interpolation among
  neighboring points. End points are treated similarly,
  but with two neighbors on one side. 

************************************************************/
double MD_d_dr(y, ax, pos)
dblptr y;
axis_ptr ax;
unsigned int pos;
{
double result, deli, deli_last, hi, hi_last, span;
dblptr r;
unsigned int i, lim, pos1, mul, posh, posl;
bool low_lim, hi_lim;

r = ax->r;
i = ax->index;
lim = ax->num_pts+1;
mul = ax->mul;
low_lim = (i == 0); hi_lim = (i == lim);
pos1 = low_lim?(pos+mul):(hi_lim?(pos-mul):pos);
posh = pos1+mul; posl = pos1-mul;
hi = r[posh] - r[pos1];
hi_last = r[pos1] - r[posl];
span = hi+hi_last;
deli = (y[posh] - y[pos1])/hi;
deli_last = (y[pos1] - y[posl])/hi_last;
result = (deli*hi_last + deli_last*hi);
if (low_lim||hi_lim){
  result += low_lim?(2.0*hi_last*(deli_last - deli)):
  (2*hi*(deli - deli_last));
}					    
result /= span;
return(result);
}
/************************************************************

  MD_FCT_Richardson(set, f, dt)

  Procedure for doing Richardson extrapolation to make a first
  order integration method into a second order method. Used
  in conjunction with MD_FCT. 

  set is the MD_species_set
  f is the defining function
  dt is the time step.

  space is allocated in sp->rho_mid if first time through with the
  set, or if set->version has changed. For speed reasons, sp->rho_mid is
  tested for NULL to see if it must be reallocated.

  Note: if not the first time through, and if size has
  changed, sp->rho_mid must be set to NULL if size has changed.

  (This allocation should probably be done in get_data_space.
  Someday it may be changed.)

************************************************************/
bool MD_FCT_Richardson(set, f, dt)
MD_FCT_species_set_ptr set;
bool (*f)();
double dt;
{
static unsigned int last_version = 0;
static MD_FCT_species_set_ptr last_set = NULL;
MD_FCT_species_set_rec_ptr rec;
MD_FCT_species_ptr spptr = NULL;
unsigned int i, size;
double hdt;
bool result;

/*
  if not the previous set (or first time) allocate space for rho_mid
  fill it with zeros
*/
if ((set != last_set)||(set->version != last_version)){
  for(rec = set->start; rec != NULL; rec = rec->next){
    spptr = rec->first;
    if (spptr->rho_mid == NULL){
      size = spptr->grid->alloc_size;
      spptr->rho_mid = get_memory(spptr->rho_mid, size, double);
    }
    for(i = 0; i < size; i++){ spptr->rho_mid[i] = 0.0;}
  }
  last_set = set;
  last_version = set->version;
}
/*
  Start the extrapolation procedure:

  if MD_FCT_step == half (Runge-Kutta half step):
  (dt):
  spptr->rho_n = spptr->rhoS_init;
  spptr->rho_np1 = spptr->S;

  if MD_FCT_step == full (Runge-Kutta full step,
  half step values are already in rhoS):
  (dt/2)
  spptr->rho_n = spptr->rhoS;
  spptr->rho_np1 = spptr->rho_mid;

  (dt):
  spptr->rho_n = spptr->rhoS_init;
  spptr->rho_np1 = spptr->rhoS;

  spptr->rhoS = 2*spptr->rho_mid - spptr->rhoS;
*/
if (MD_FCT_step == half){
  for(rec = set->start; rec != NULL; rec = rec->next){
    spptr = rec->first;
    spptr->rho_n = spptr->rhoS_init;
    spptr->rho_np1 = spptr->rhoS;
  }
  result = f(set, dt);
  return(result);
}
else if (MD_FCT_step == full){
  hdt = .5*dt;
  for(rec = set->start; rec != NULL; rec = rec->next){
    spptr = rec->first;
    spptr->rho_n = spptr->rhoS;
    spptr->rho_np1 = spptr->rho_mid;
  }
  result = f(set, hdt);
  if (!result) { return(FALSE); }
  for(rec = set->start; rec != NULL; rec = rec->next){
    spptr = rec->first;
    spptr->rho_n = spptr->rhoS_init;
    spptr->rho_np1 = spptr->rhoS;
  }
  result = f(set, dt);
  if (!result) { return(FALSE); }
  for(rec = set->start; rec != NULL; rec = rec->next){
    spptr = rec->first;
    size = spptr->grid->alloc_size;
    for(i=0; i< size; i++){
      spptr->rhoS[i] = 2.0*spptr->rho_mid[i] - spptr->rhoS[i];
    }
  }
  return(TRUE);
} else {
  return(FALSE);
}
}
/************************************************************

  MD_tri_diag(A, u, r, N_int, ssteps, new)

  Solves problem Au=r, with u0...uN_int+1 in u[0]...u[N_int+1]. 
  r must be of length N_int+2.

  tri-diagonal matrix solver using sweep method.
  matrix A contains all three vectors of length N_int+2
  where N_int is the number of "internal" points (MD form).
  stored as:

b0 c0
a1 b1 c1
        ........
	 aN_int   bN_int   cN_int

	          aN_int+1 bN_int+1 


where a1...aN_int+1 ==> A[1]...A[N_int+1], 
b0...bN_int ==> A[N_int+2]...A[2N_int+3],
and c0...cN_int-1 ==> A[2N_int+4]... A[3N_int+4]

so that A must be at least 3N_int+5 in size.

A is not overwritten.

r and u are in MD form. Thus, ssteps is the MD index step value
corresponding to the axis along which the tri-diagonal matrix
is being applied. r and u are pointers to the left edge of 
the axis; such as &(rho[sl]).

r and u can point to the same space.

Previous elimination elements are saved. Parameter "new" determines if
elimination is to be performed on A, or if previous values are to be
used again. Re-use (new == false) saves two floating operations per
step.

Returns true if successful. 

************************************************************/
bool MD_tri_diag(A, u, r, N_int, ssteps, new)
dblptr A, u, r;
unsigned int N_int;
bool new;
{
static dblptr gamma = NULL, beta = NULL;
static unsigned int last_N = 0, s;
double bet, bet_, gam, aa, unext;
dblptr a, b, c;
unsigned int i, N;

N = N_int+2;
if (N > last_N) {
  gamma = get_memory(gamma, (N+1), double);
  beta = get_memory(beta, (N+1), double);
  last_N = N;
  new = true;
}

a = A;
b = a+N;
c = b+N;

bet = b[0];
if (new){
  for(i = 1; ; i++){
    if (bet == 0.0){
      fprintf(stderr,"[singularity in MD_tri_diag...terminating]\n");
      return(false);
    }
    beta[i] = (bet_ = 1.0/bet);
    if (i == N) { break;}
    gamma[i] = (gam = c[i-1]*bet_);
    bet = b[i] - a[i]*gam;
  }
}
aa = 0.0; s = 0;
for(i=1; ; i++){
  u[s] = ((r[s] - aa)*beta[i]);
  if (i == N) {break;}
  aa = a[i]*u[s];
  s += ssteps;
}
unext = u[s];
for(i=(N-1); i>=1; i--){
  s -= ssteps;
  unext = (u[s] -= gamma[i]*unext);
}
return(true);
}
/************************************************************

  MD_dummy_bc_fn()

  Dummy routine to return immediately. Placeholder
  for a user-provided routine which will affect
  S, V, L_char, v, etc.

************************************************************/
static void MD_dummy_bc_fn(bc)
     MD_FCT_bc *bc;
{
  return;
}
/************************************************************

  Routines to scan input command line

  MD_register bool(c,&b) to register flag c to set bool b
  MD_register string(c,&s) to register flag c to set string s
  MD_register_double(c,&d) to register flag c to set double d
  MD_register_uint(c,&u) to register flag c to set unsigned int u

  MD_scan_command_line(argc,argv) to scan command line for
  flags preceded by '-', or an input file name with no '-'. 

************************************************************/
typedef struct MD_input_param{
  char type;
  char flag;
  void *data;
} MD_input_param;

typedef MD_input_param *MD_input_param_ptr;

static MD_input_param MD_inputs[MD_N_params];

int MD_param_cnt = 0;

void MD_register_bool(c, b)
     char c;
     bool *b;
{
  if (MD_param_cnt >= MD_N_params) {
    fprintf(stderr,"MD_register_bool: Too many parameters\n");
    exit(1);
  }
  MD_inputs[MD_param_cnt].type = 'b';
  MD_inputs[MD_param_cnt].flag = c;
  if (b == NULL){
    fprintf(stderr,"MD_register_bool: Invalid address for flag %c\n",c);
    exit(1);
  }
  MD_inputs[MD_param_cnt].data = (void *)b;
  *b = FALSE;
  MD_param_cnt++;
  return;
}

void MD_register_string(c, s)
     char c;
     string *s;
{
  if (MD_param_cnt >= MD_N_params) {
    fprintf(stderr,"MD_register_string: Too many parameters\n");
    exit(1);
  }
  MD_inputs[MD_param_cnt].type = 's';
  MD_inputs[MD_param_cnt].flag = c;
  MD_inputs[MD_param_cnt].data = (void *)s;
  MD_param_cnt++;
  return;
}

void MD_register_uint(c, i)
     char c;
     unsigned int *i;
{
  if (MD_param_cnt >= MD_N_params) {
    fprintf(stderr,"MD_register_uint: Too many parameters\n");
    exit(1);
  }
  MD_inputs[MD_param_cnt].type = 'u';
  MD_inputs[MD_param_cnt].flag = c;
  MD_inputs[MD_param_cnt].data = (void *)i;
  MD_param_cnt++;
  return;
}

void MD_register_double(c, d)
     char c;
     double *d;
{
  if (MD_param_cnt >= MD_N_params) {
    fprintf(stderr,"MD_register_double: Too many parameters\n");
    exit(1);
  }
  MD_inputs[MD_param_cnt].type = 'd';
  MD_inputs[MD_param_cnt].flag = c;
  MD_inputs[MD_param_cnt].data = (void *)d;
  MD_param_cnt++;
  return;
}

void MD_scan_command_line(c,v)
     int c;
     char **v;
{
  char **last_ptr = &(v[c-1]); /* pointer to last argument */
  char **endptr, *chptr, *opt_ptr, *strname = NULL, *s= NULL, **sptr;
  int i,j;
  unsigned int u = 0,*uptr;
  double d = 0.0, *dptr;
  bool *bptr, found_it;

  endptr = &chptr;
  while(++v <= last_ptr){                 // scan through input parameters
    if((opt_ptr = *v)[0] == '-'){         // look for a preceding '-'
      for(i=1; opt_ptr[i] != '\0'; i++){  // look for a "bunch" of parameters
                                          // with one common '-'
	found_it = FALSE;
	for(j=0; j<MD_param_cnt; j++){    // scan through all registered 
                                          // parameters and look for this one
	  if(opt_ptr[i] ==  MD_inputs[j].flag){
	    switch(MD_inputs[j].type)
	      {
	      case 'b':                   // it's a bool! set the flag
		bptr = (bool *)MD_inputs[j].data;
		*bptr = TRUE;
		found_it = TRUE;
		break;
	      case 'u':                   // it's an unsigned int! set it
		if(v <= last_ptr){
		  u = (unsigned int)strtoul(*(++v),endptr,0);
		  if(**endptr != '\0'){
		    fprintf(stderr,"MD_scan_command_line: bad value (flag %c): %s\n",opt_ptr[i],*v);
		    exit(1);
		  }
		  uptr = (unsigned int *)MD_inputs[j].data;
		  *uptr = u;
		}
		found_it = TRUE;
		break;
	      case 'd':                    // it's a double! set it
		if (v <= last_ptr){
		  d = strtod(*(++v),endptr);
		  if (**endptr != '\0') {
		    fprintf(stderr,"MD_scan_command_line: bad parameter (flag %c): %s\n",opt_ptr[i],*v);
		    exit(1);
		  }
		  dptr = (double *)MD_inputs[j].data;
		  *dptr = d;
		}
		found_it = TRUE;
		break;
	      case 's':                   // it's a string! set it
		if (v <= last_ptr) {
		  sptr = (char **)MD_inputs[j].data;
		  s = strdup(*(++v));
		  if (s == NULL){
		    fprintf(stderr,"MD_scan_command_line: Insufficient memory at parameter %c %s\n",opt_ptr[i],*v);
		    exit(1);
		  }
		  *sptr = s;
		}
		found_it = TRUE;
		break;
	      default:                   // couldn't find the type...should not
                                         // be possible to get here because all
                                         // types are accounted for by the
                                         // registering functions
		fprintf(stderr,"MD_scan_command_line: bad option %c\n",(*v)[i]);
		exit(1);
		break;
	      }
	  }
	}
	if (!found_it){                  // command line flag wasn't registered
	  fprintf(stderr,"MD_scan_command_line: option %c not registered\n",(*v)[i]);
		exit(1);
		break;
	      }
	if (opt_ptr != *v) {break;}   // this flag took a parameter, so move on
      }
    } else {
      if (!MD_FCT_INPUT_FILE_FLAG){ // One parameter without a flag is allowed;
                                    // assumed to be the name of an input file 
	MD_FCT_INPUT_FILE_NAME = *v;
	if (MD_FCT_INPUT_FILE_NAME == NULL){
	  fprintf(stderr,"MD_scan_command_line: bad input file name %s\n",(*v));
	  exit(1);
	}
	MD_FCT_INPUT_FILE_FLAG = TRUE;
      } else {                  // Somebody passed a second unflagged parameter
	fprintf(stderr,"MD_scan_command_line: unknown parameter %s\n",(*v));
      }
    }
  }
/************************************************************
  Now that everything is processed, set up directory for 
  input/output data. Begin by making sure MD_DATADIR exists.
************************************************************/
  strname = strdup(MD_FCT_OUTPUT_FILE_NAME);
  s = NULL;
  s = get_memory(s,(strlen(MD_DATADIR)+strlen(strname)+1+strlen("mkdir -p ")),
		 char);
  if (s == NULL){
    fprintf(stderr,"MD_scan_command_line: Insufficient memory\n");
    exit(1);
  }
  strcpy(s,"mkdir -p ");
  strcat(s,MD_DATADIR);
  if ( system(s) == -1){
    fprintf(stderr,"system call: %s failed\n",s);
  }
  strcpy(s,MD_DATADIR);
  if (s[strlen(s)-1] != '/'){
    strcat(s,"/");
  }
  strcat(s,strname);
  MD_FCT_OUTPUT_FILE_NAME = s;
  MD_FCT_OUTPUT_FILE = fopen(MD_FCT_OUTPUT_FILE_NAME, "w");
  if (MD_FCT_OUTPUT_FILE == NULL){
    fprintf(stderr,"MD_scan_command_line: unable to open output file %s\n",MD_FCT_OUTPUT_FILE_NAME);
    exit(1);
  }
  fprintf(stderr,"MD_scan_command_line: directing output to %s\n", MD_FCT_OUTPUT_FILE_NAME);

/************************************************************

  Now set up input file

************************************************************/
  if (MD_FCT_INPUT_FILE_FLAG){  // If an input file was specified, try to
                                // open it. Look in MD_DATADIR (which defaults
                                // to "." but can be modified)
    strname = strdup(MD_FCT_INPUT_FILE_NAME);
    s = NULL;
    s = get_memory(s,(strlen(MD_DATADIR)+strlen(strname)+1),char);
    if (s == NULL){
      fprintf(stderr,"MD_scan_command_line: Insufficient memory\n");
      exit(1);
    }
    strcpy(s,MD_DATADIR);
    if (s[strlen(s)-1] != '/'){
      strcat(s,"/");
    }
    strcat(s,strname);
    MD_FCT_INPUT_FILE_NAME = strdup(s);
    MD_FCT_INPUT_FILE = NULL;
    MD_FCT_INPUT_FILE = fopen(MD_FCT_INPUT_FILE_NAME,"r");
    if (MD_FCT_INPUT_FILE == NULL){
      fprintf(stderr,"MD_scan_command_line: unable to open file \"%s\"\n",(MD_FCT_INPUT_FILE_NAME));
      exit(1);
    }
  }
  free(s); free(strname);
  return;    
}
    



