/************************************************************

  MD_FCT_utils.h

  Created by Edward A. Richley (richley@mailaps.org)

  This file is subject to the terms and conditions of the 
  GNU General Public License. 


************************************************************/
#ifndef MD_FCT_utils_flag
#define MD_FCT_utils_flag

#include "MD_FCT_defs.h"

extern string MD_FCT_convection_names[];
extern FILE *MD_FCT_OUTPUT_FILE;
extern string MD_FCT_OUTPUT_FILE_NAME;
extern FILE *MD_FCT_INPUT_FILE;
extern string MD_FCT_INPUT_FILE_NAME;
extern bool MD_FCT_INPUT_FILE_FLAG;
extern string MD_DATADIR;

extern MD_FCT_flow_step MD_FCT_step;
/************************************************************

  MD_FCT_new_axis(NAME,UNITS,N,grid, Length, fn)

  Makes a new axis named "NAME" of N cells, inserts it in
  grid, and uses fn() to define the positions/areas/volumes.
  Returns a pointer to it.

************************************************************/
extern axis_ptr MD_FCT_new_axis(string, string, unsigned int, geometry_ptr, double, void());
/************************************************************
  
  MD_FCT_new_species_set(name)

  Makes a new species set named "name". Returns a pointer to it.

************************************************************/
extern MD_FCT_species_set_ptr MD_FCT_new_species_set();
/************************************************************
  
  MD_FCT_new_species(NAME, UNITS, set, grid)

  Makes a new species named "NAME", inserts it in species set
  "set", allocates storage according to dimensions from
  "grid". "UNITS" is a string containing the physical units of
  the species. Returns a pointer to it.

************************************************************/
extern MD_FCT_species_ptr MD_FCT_new_species(string, string, MD_FCT_species_set_ptr, geometry_ptr);
/************************************************************
  
  MD_get_axptr(NAME, grid)

  getis axis ptr for axis with name NAME on grid

************************************************************/
extern axis_ptr MD_get_axptr();
/************************************************************

  MD_new_reaction(NAME, rlist, set, splist, fn)

  generate an entry in rlist->start for a new reaction 
  with name NAME, and  rate function fn. 
  set is the species set in which reactants will be found.
  splist is a NULL-terminated list of names  of reactants.
  deltas is a vector of net changes for each reactant in splist.
  ratefn(reaction,set) = pointer to function defining net reaction rate.
  Returns pointer to reaction.

************************************************************/
MD_reaction_ptr MD_new_reaction();
/************************************************************

  MD_new_reaction_list(name)

  Allocate space for a new reaction list named name.
  Begins linked list to set of reactions.

************************************************************/
MD_reaction_list_ptr MD_new_reaction_list();
/************************************************************

  MD_get_spptr(NAME, SET)

  gets ptr to species with name NAME in SET

************************************************************/
extern MD_FCT_species_ptr MD_get_spptr();
/************************************************************

  MD_integral_dV()

  Integrates vector y over "active" region of axis
  ax (1 thru num_pts) dVolume. Simple dot product
  to be consistent with finite difference representation.

************************************************************/
extern double MD_integral_dV(dblptr, axis_ptr, geometry_ptr);
/************************************************************

  MD_integral_dVol()

  Integrates array y over "active" region of axis
  ax (1 thru num_pts) dVolume. Simple dot product
  to be consistent with finite difference representation.

************************************************************/
extern double MD_integral_dVol(dblptr, axis_ptr, geometry_ptr);
/************************************************************
  
  MD_integral_dr()

  Integrates vector y over "active" region of axis
  ax (halfway between r[0] and r[1] thru halfway
  between r[num_pts] and r[num_pts+1]) dr. 
  Implements Hermite Cubic Quadrature algorithm
  as described by Kahaner, Moler, and Nash, chs. 4
  and 5. End points have strange coefficients because
  integration is not from ax->r[0] to ax->r[num_pts+1]. 

************************************************************/
extern double MD_integral_dr();
/************************************************************

  MD_get_position()

  gets current position pointed to by all axes in "grid"

************************************************************/
extern unsigned int MD_get_position();
/************************************************************

  MD_dy_dx(y,ax)

  gets derivative estimate for vector "y"
  with respect to axis "ax" at position 
  set by ax->index.

  Uses MD_dY_dx()

  computation is done by quadratic interpolation among
  neighboring points. End points are treated similarly,
  but with two neighbors on one side. 

  Any axis adjustment should be done outside.

************************************************************/
/*extern double MD_dy_dx();*/
/************************************************************

  MD_dY_dx(y,x, N, i)

  gets derivative estimate for vector "Y"
  with respect to vector "x" of length N
  at position i 

  computation is done by quadratic interpolation among
  neighboring points. End points are treated similarly,
  but with two neighbors on one side. 

  Any axis adjustment should be done outside.

************************************************************/
extern double MD_dY_dx();
/************************************************************

  MD_d_dr(y,ax,pos)

  gets derivative estimate for array "y"
  with respect to axis "ax" at position "pos".
  "pos" is probably derived from a call to MD_get_position.
  The current index of "ax" must correspond to "pos". 

  This way, the "species" never enters the calculation,
  so it can be used for non-species arrays.

  computation is done by quadratic interpolation among
  neighboring points. End points are treated similarly,
  but with two neighbors on one side. 

************************************************************/
extern double MD_d_dr();
/************************************************************

  MD_output_all_rho(set)

  Routine to print all concentration data for a species set.
  Prints in format compatible with gnuplot.
  Prints into directory MD_DATADIR under file named
  species".dat". Old data is overwritten. 

************************************************************/
extern void MD_output_all_rho(MD_FCT_species_set_ptr);
/************************************************************

  MD_output_vel(sp, vel)

  Routine to print velocity data for a species. 
  
  Prints in format compatible with gnuplot.
  Prints into directory MD_DATADIR under file named
  sp->name.vel->name".dat". Old data is overwritten. 
  

************************************************************/
extern void MD_output_vel(MD_FCT_species_ptr, MD_FCT_vel_list_rec_ptr);
/************************************************************

  MD_output_all_vel(sp)

  Routine to print all velocity data for a species. 
  Prints in format compatible with gnuplot.
  Prints into directory MD_DATADIR under file named
  species_name.velocity_name".dat". Old data is overwritten. 

************************************************************/
extern void MD_output_all_vel(MD_FCT_species_ptr);
/************************************************************

  MD_FCT_print_all_rho()

  Routine to print values of rho0 for a species set. 

************************************************************/
extern void MD_FCT_print_all_rho(MD_FCT_species_set_ptr);
/************************************************************

  MD_FCT_read_all_rho()

  Routine to read values of rho0 for a species set. 

************************************************************/
extern void MD_FCT_read_all_rho();
/************************************************************

  MD_FCT_print_vel()

  Routine to print values of velocity.

************************************************************/
extern void MD_FCT_print_vel();
/************************************************************

  MD_FCT_rect(ax, Length)

  Defines axis "ax" to have rectangular dimensions with 
  number of cells defined by "Length".

************************************************************/
extern void MD_FCT_rect();
/************************************************************

  MD_FCT_cyl(ax, Length)

  Defines axis "ax"to have cylindrical dimensions with
  number of cells defined by "Length".

************************************************************/
extern void MD_FCT_cyl();
/************************************************************

  MD_FCT_sphere(ax, Length)

  Defines axis "ax"to have spherical dimensions with
  number of cells defined by "Length".

************************************************************/
extern void MD_FCT_sphere();
/************************************************************

  Axis definition routines for polar coordinates,
  with r and theta axes.

  MD_polar_r()

  Axis initialization routine. 
  Dimensions radius axes for use with
  polar coordinates (r, theta).

            0  ........................Length
            |                            |
Volume:   0 | 1 | 2 | .... | i | ... | N | N+1 |  
r:          0   1   2     i-1 i+1   N-1  N
Area:       0   1   2     i-1 i+1   N-1  N

equally spaced cells with size Length/N.

  Makes new Area and Volume vectors, depending on 
  delta_theta.

************************************************************/
extern void MD_polar_r();
/************************************************************

  MD_polar_theta()

  Axis initialization routine. 
  Dimensionins azimuth axis
  for use with polar coordinates (r, theta)

            0  ........................Aperture
            |                            |
Volume:   0 | 1 | 2 | .... | i | ... | N | N+1 |  
r:          0   1   2     i-1 i+1   N-1  N
Area:       0   1   2     i-1 i+1   N-1  N

equally spaced cells with size delta_theta = Aperture/N.  Area and
Volume depend on r position.  Makes Area and Volume vectors, depending
on r position, as given by r_index_ptr.



************************************************************/
extern void MD_polar_theta();
/************************************************************

  MD_FCT_set_bc(spe, ax, edge, S, V, location)

  Sets "edge" (left or right) boundary conditions of
  species "spe" for axis "ax" on grid "grid" to S, V.
  Thus, for (edge == left) spe->bc[i].Sl = S and
  spe->bc[i].Vl = V where i is the index of axis "ax"
  and "ax" is given by string name.

************************************************************/
extern void MD_FCT_set_bc();
/************************************************************

  MD_FCT_set_bc_fn(spe, ax, edge, fn)

  Sets boundary condition at "edge" of species "spe" for
  axis "ax" to be: fn

  If (fn == NULL), no change is made. This allows
  the default function "MD_dummy_bc_fn()" to be kept.
  
  If function (fn != NULL), it is assumed that the
  provided function will affect S, V, L_char, and v_char
  so as to be useful. As so, these are unchanged.

  If it hasn't already been set, OFFSET is set to correspond 
  to local conditions as a default. 

  No other parameters are affected (location, S, V,
  L_char, v) so these must be specified
  by a call to MD_FCT_set_bc(), or directly modified
  outside fn().

************************************************************/
extern void MD_FCT_set_bc_fn(MD_FCT_species_ptr, string, enum boundary, void (*)());
/************************************************************

  MD_get_spptr(NAME, SET)

  Gets pointer to species with name NAME from set SET.
  Returns NULL if not found.

************************************************************/
extern MD_FCT_species_ptr MD_get_spptr();
/************************************************************

  MD_get_axptr(NAME, grid)

  Returns pointer to axis NAME in grid. Returns NULL
  if not found.

************************************************************/
extern axis_ptr MD_get_axptr();
/************************************************************

  MD_FCT_create_velocity(spe, ax, convection, name, units)

  Creates velocity for species pointer "spe" on axis name "ax" 
  with convection type "convection" name "name" and units "units".

************************************************************/
extern MD_FCT_vel_list_rec_ptr MD_FCT_create_velocity(MD_FCT_species_ptr, string, MD_FCT_convection, string, string);
/************************************************************

  MD_FCT_Richardson(set, f, dt)

  Procedure for doing Richardson extrapolation to make a first
  order integration method into a second order method.

************************************************************/
extern bool MD_FCT_Richardson(MD_FCT_species_set_ptr, bool(), double);
/************************************************************

  MD_tri_diag(A, u, r, N_int, ssteps, new)

  tri-diagonal matrix solver using sweep method.
  matrix A contains all three vectors of length N_int+2
  where N_int is the number of "internal" points (MD form).
  stored as:

  Solves problem Au=r, with u0...uN_int+1 in u[0]...u[N_int+1]. 
  r must be of length N_int+2.

b0 c0
a1 b1 c1
        ........
	 aN_int   bN_int   cN_int

	          aN_int+1 bN_int+1 


where a1...aN_int+1 ==> A[1]...A[N_int+1], 
b0...bN_int ==> A[N_int+2]...A[2N_int+3],
and c0...cN_int-1 ==> A[2N_int+4]... A[3N_int+4]

so that A must be at least 3N_int+5 in size.

A is not overwritten.

r and u are in MD form. Thus, ssteps is the MD index step value
corresponding to the axis along which the tri-diagonal matrix
is being applied. r and u are pointers to the left edge of 
the axis; such as &(rho[sl]).

r and u can point to the same space.

Previous elimination elements are saved. Parameter "new" determines if
elimination is to be performed on A, or if previous values are to be
used again. Re-use (new == false) saves two floating operations per
step.

Returns true if successful. 

************************************************************/
extern bool MD_tri_diag();

/************************************************************

  Routines to scan input command line

  MD_register bool(c,&b) to register flag c to set bool b
  MD_register string(c,&s) to register flag c to set string s
  MD_register_double(c,&d) to register flag c to set double d
  MD_register_uint(c,&u) to register flag c to set unsigned int u

  MD_scan_command_line(argc,argv) to scan command line for
  flags preceded by '-', or an input file name with no '-'. 

************************************************************/
void MD_register_bool(char, bool *);
void MD_register_string(char, string *);
void MD_register_uint(char, unsigned int *);
void MD_register_double(char, double *);
void MD_scan_command_line(int,char **);

#endif

