/************************************************************

  MD_tabletop.c

  Example program for using MD_FCT routines. 

  Solves:   d(rho)/dt + deldot(rho vel) = -rho/tau

  in one or two dimensions on a rectangular grid Nx by Ny.

  Velocity is constant at (1.0 1.0). Initial conditions
  are a "tabletop" or height 2.0 over an ambient of
  height 1.0.

  Created by Edward A. Richley (richley@mailaps.org)

  This file is subject to the terms and conditions of the 
  GNU General Public License. 


************************************************************/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "basics.h"
#undef pi
#include "MD_FCT_defs.h"
#include "MD_FCT_utils.h"
#include "MD_FCT.h"
#include "MD_plplot.h"


#define Nx 50
#define Ny 50
#define XLENGTH 10.0
#define YLENGTH 10.0
#define XVEL 1.0
#define YVEL 1.5

#define TABLETOP 2.0
#define FLOOR 1.0

static bool oned = false; /* bit to tell if it's a 1d or 2d problem */
static double tau = 500.0; /* Time constant for source term */

static unsigned int PXLOW = Nx/4, PXHI = Nx/2, PYLOW = Ny/4, PYHI = Ny/2;


/************************************************************

  ideal(sp)

  Ideal tabletop convection solution after time
  MD_Total_time. Writes solution into sp->rho0.

************************************************************/
void ideal(sp)
MD_FCT_species_ptr sp;
{
unsigned int i, p, s, alloc_size, Nymax, pxlow, pxhi, pylow, pyhi;
bool test;
double fac;

fac = exp(-MD_Total_time/tau);
alloc_size = (Nx+2)*(Ny+2);

pxlow = PXLOW + (int)(Nx*XVEL*MD_Total_time/XLENGTH +.5);
pxhi = PXHI + (int)(Nx*XVEL*MD_Total_time/XLENGTH +.5);
pylow = PYLOW + (int)(Ny*YVEL*MD_Total_time/YLENGTH +.5);
pyhi = PYHI + (int)(Ny*YVEL*MD_Total_time/YLENGTH +.5);

if (oned) {Nymax = 0;} else {Nymax = (Ny+1);} 
for(i=0; i<=Nymax; i++){
    s = i*(Nx+2);
    for(p=0; p<=(Nx+1); p++){
      test = ((p >= pxlow)&(p <= pxhi)) ;
      test &= (oned)?true:((i >= pylow)&(i <= pyhi));
      if (test) {
	(sp->rho0[p+s] = TABLETOP*fac);
      } else {
	(sp->rho0[p+s] = FLOOR*fac);
      }
    }
  }

}

/************************************************************

  sources()

  get velocities from rho. Get rhoS from rhoS_init.
  In this example, velocity is constant, and is only
  computed the first time through.

************************************************************/
bool sources(set, time)
double time;
MD_FCT_species_set_ptr set;
{

MD_FCT_species_ptr sp;
unsigned int i, p, s, v, alloc_size, Nymax;
static bool First_Time = true;
double fac, decay;

sp = MD_get_spptr("rho",set);
alloc_size = oned?(Nx+2):((Nx+2)*(Ny+2));
/*
  get velocities if first time
*/
if (First_Time){
  Nymax = (oned)?0:(Ny+1);
  for(i=0; i<=Nymax; i++){
    s = i*(Nx+2);
    for(p=0; p<=(Nx+1); p++){
      sp->vel->velocity[p+s] = XVEL;      /* set vx */
      sp->vel->next->velocity[p+s] = YVEL; /* set vy */
    }
  }
  First_Time = false;
}
/*
  update boundaries
*/
fac = FLOOR*exp(-MD_Total_time/tau);
MD_FCT_set_bc(sp, "x", left, 0.0, fac, local);
MD_FCT_set_bc(sp, "x", right, 0.0, fac, local);
if (!oned) {
  MD_FCT_set_bc(sp, "y", left, 0.0, fac, local);
  MD_FCT_set_bc(sp, "y", right, 0.0, fac, local);
}

/*
  d(rho)/dt + deldot(rho vel) = -rho/tau, 

  so source term is approximated by second order
  implicit method:

  rhoS = rhoS_init - gamma*rhoS

  where gamma = (dt/tau)*(1+(1/4)*(dt/tau))

*/
decay = time/tau;
decay = decay*(1.0+.25*decay);
decay = 1.0/(1.0+decay);
/*
  apply source terms
*/
for(p=0; p<alloc_size; p++){
  sp->rhoS[p] = sp->rhoS_init[p]*decay;
}
/* adjust characteristic source time for MD_Flow */
 MD_Source_dt = tau; 
 return(TRUE);
}
/************************************************************

  Initialize()

  Initialization routine to set up "tabletop" test
  initial conditions

************************************************************/
void Initialize(set)
MD_FCT_species_set_ptr set;
{

unsigned int i, p, s, alloc_size, Nymax;
MD_FCT_species_ptr sp;
bool test;

alloc_size = (Nx+2)*(Ny+2);
sp = set->start->first;
if (oned) {Nymax = 0;} else {Nymax = (Ny+1);} 
for(i=0; i<=Nymax; i++){
    s = i*(Nx+2);
    for(p=0; p<=(Nx+1); p++){
      test = ((p <= PXHI)&(p>=PXLOW)) ;
      test &= (oned)?true:((i <= PYHI)&(i >= PYLOW));
      if (test) {
	(sp->rho0[p+s] = TABLETOP);
	(sp->rhoS[p+s] = TABLETOP);
	(sp->rho1[p+s] = TABLETOP);
      } else {
	(sp->rho0[p+s] = FLOOR);
	(sp->rhoS[p+s] = FLOOR);
	(sp->rho1[p+s] = FLOOR);
      }
    }
  }

}

/************************************************************

  main()

************************************************************/ 
main(){
  geometry_ptr mygrid;
  double deltat, dt, max_eps;
  unsigned int num_dim, i,j,p,s,v,r,alloc_size;
  MD_FCT_species_ptr sp;
  axis_ptr *ax;
  MD_FCT_species_set_rec_ptr  *rec;
  MD_FCT_species_set_ptr myset = NULL;
  MD_FCT_vel_list_rec_ptr vp;
  
/* 
  Initialize species set "MySet" and grid "MyGrid"
*/
  myset = MD_FCT_new_species_set("MySet");
  mygrid = MD_FCT_new_grid("mygrid");
/*
  Define axes for MyGrid as rectangular of length 10
  in each direction with Nx x points, and Ny y points
*/

  MD_FCT_new_axis("x", "M", Nx, mygrid, XLENGTH, MD_FCT_rect);
  if (!oned){
    MD_FCT_new_axis("y", "M", Ny, mygrid, YLENGTH, MD_FCT_rect);
  }
/*
  Define a species called "rho", defined on MyGrid,
  and insert it in set MySet. Also, set the
  background color to blue.
*/
  sp = MD_FCT_new_species("rho", "M-3",  myset, mygrid);
  sp->enable_minimum = TRUE;
  sp->minimum = 0.0;
  MD_FCT_create_velocity(sp, "x", conservation, "vx", "M/s");
  MD_FCT_create_velocity(sp, "y", conservation, "vy", "M/s");
/*
  Initialize the concentration data
*/
  Initialize(myset);
/*
  plot the initial data
*/
  plot_all_rho(myset,red);
/*
  Flow MySet for time 2.0, with velocities and sources
  defined by procedure "sources()" , and plot the results
*/
  fprintf(stderr,"done plotting...\n");
  MD_Flow(myset, sources, 2.0);

  fprintf(stderr,"done flowing once...\n");
  plot_all_rho(myset,red);
  fprintf(stderr,"done plotting second...\n");
/*
  Flow MySet for another 2.0, change the viewing angle, 
  and plot the results
*/
  MD_Flow(myset, sources, 2.0);
  fprintf(stderr,"done flowing twice...\n");
/*sp->ViewAngle = "[90 0]";*/
  plot_all_rho(myset,red);

/*
  Compute the ideal solution for this time,
  plot it on top of the last plot
*/
/*ideal(sp);
plot_all_rho(myset,green);*/
}












